import copy
import hashlib
import sys
import numpy as np
import pandas as pd

from typing import Any
from pathlib import Path

from mirp._images.base_image import BaseImage
from mirp.settings.generic import SettingsClass

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class GenericImage(BaseImage):

    def __init__(
            self,
            image_data: None | np.ndarray,
            translation: None | tuple[float, ...] = None,
            rotation_angle: None | float = None,
            noise_iteration_id: None | int = None,
            noise_level: None | float = None,
            interpolated: bool = False,
            interpolation_algorithm: None | str = None,
            discretisation_method: None | str = None,
            discretisation_bin_number: None | int = None,
            discretisation_bin_width: None | float = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Image data. Note that image_data is explicitly defined as this prevents IDE warnings. However,
        # the attribute itself is set using set_voxel_grid. Also, note that we explicitly copy image_data because
        # otherwise we may end up changing objects by reference, which is not the expected behaviour.
        self.set_voxel_grid(copy.deepcopy(image_data)) if image_data is not None else None

        # Perturbation-related settings that are set during interpolate.
        self.translation = translation
        self.rotation_angle = rotation_angle
        self.noise_iteration_id = noise_iteration_id
        self.noise_level = noise_level

        # Interpolation-related settings
        self.interpolated = interpolated
        self.interpolation_algorithm = interpolation_algorithm

        # Discretisation-related settings
        self.discretisation_method = discretisation_method
        self.discretisation_bin_number = discretisation_bin_number
        self.discretisation_bin_width = discretisation_bin_width

        # Slice identifiers.
        self.slice_id: None | int = None

        # Image underwent IBSI compliant processing (default True: it is easier to identify exceptions).
        self.ibsi_compliant = True

    def update_from_template(self, template):
        if not isinstance(template, GenericImage):
            raise TypeError(
                f"The new class object should inherit from the template provided by the \"image\" object, which is "
                f"expected to inherit from GenericImage. Found: {type(template)}")

        # NOTE: image_data is not set automatically.

        # Attributes from BaseImage
        self.modality = template.modality
        self.image_origin = copy.deepcopy(template.image_origin)
        self.image_orientation = copy.deepcopy(template.image_orientation)
        self.image_spacing = copy.deepcopy(template.image_spacing)
        self.image_dimension = copy.deepcopy(template.image_dimension)
        self.sample_name = copy.deepcopy(template.sample_name)
        self.object_metadata = copy.deepcopy(template.object_metadata)

        # Attributes from GenericImage
        self.separate_slices = template.separate_slices
        self.translation = copy.deepcopy(template.translation)
        self.rotation_angle = copy.deepcopy(template.rotation_angle)
        self.noise_iteration_id = template.noise_iteration_id
        self.noise_level = template.noise_level
        self.interpolated = template.interpolated
        self.interpolation_algorithm = template.interpolation_algorithm
        self.discretisation_method = template.discretisation_method
        self.discretisation_bin_number = template.discretisation_bin_number
        self.discretisation_bin_width = template.discretisation_bin_width
        self.ibsi_compliant = template.ibsi_compliant

    def promote(self):
        from mirp._images.ct_image import CTImage
        from mirp._images.pet_image import PETImage
        from mirp._images.mr_image import MRImage
        from mirp._images.mr_adc_image import MRADCImage
        from mirp._images.mr_dce_image import MRDCEImage
        from mirp._images.rtdose_image import RTDoseImage
        from mirp._images.digital_xray_image import DXImage, CRImage, MGImage

        if self.modality == "ct":
            image = CTImage(image_data=self.image_data)
        elif self.modality in ["pet", "pt"]:
            image = PETImage(image_data=self.image_data)
        elif self.modality in ["mr", "mri"]:
            image = MRImage(image_data=self.image_data)
        elif self.modality in ["adc"]:
            image = MRADCImage(image_data=self.image_data)
        elif self.modality in ["dce"]:
            image = MRDCEImage(image_data=self.image_data)
        elif self.modality in ["rtdose"]:
            image = RTDoseImage(image_data=self.image_data)
        elif self.modality in ["cr"]:
            image = CRImage(image_data=self.image_data)
        elif self.modality in ["dx"]:
            image = DXImage(image_data=self.image_data)
        elif self.modality in ["mg"]:
            image = MGImage(image_data=self.image_data)
        elif self.modality == "generic":
            return self
        else:
            raise ValueError(f"The imaging modality was not recognised: {self.modality}")

        # Set remaining attributes.
        image.update_from_template(template=self)

        return image

    def get_slices(
            self,
            slice_number: None | int | list[int] = None
    ) -> None | Self | list[Self]:

        image_list = []
        return_list = True

        if slice_number is None:
            slice_number = list(range(self.image_dimension[0]))
        elif isinstance(slice_number, int):
            return_list = False
            slice_number = [slice_number]

        for current_slice_id in slice_number:
            # Check if slice is not empty and falls within the allowed range.
            if self.is_empty():
                continue
            if not 0 <= current_slice_id < self.image_dimension[0]:
                continue

            slice_image = self.copy(drop_image=True)
            slice_image.image_origin = tuple(self.to_world_coordinates(x=np.array([current_slice_id, 0, 0])))
            slice_image.slice_id = current_slice_id
            slice_image.set_voxel_grid(voxel_grid=self.get_voxel_grid()[current_slice_id:current_slice_id + 1, :, :])

            image_list += [slice_image]

        if len(image_list) == 0:
            return None
        elif return_list:
            return image_list
        else:
            return image_list[0]

    def is_empty(self) -> bool:
        return self.image_data is None

    def set_voxel_grid(self, voxel_grid: np.ndarray):
        self.image_data = voxel_grid
        self.image_dimension = tuple(voxel_grid.shape)
        self.update_image_data()

    def get_voxel_grid(self) -> None | np.ndarray:
        return self.image_data

    def update_image_data(self):
        pass

    def update_separate_slices(self, x: bool):
        self.separate_slices = x

    def show(self, mask=None, slice_id=None):  # pragma: no cover
        import matplotlib.pyplot as plt
        from mirp._images.utilities import InteractivePlot

        if self.is_empty():
            return

        if mask is not None:
            # Ensure that the mask is registered to the image.
            mask = mask.copy()
            mask.register(
                image=self,
                spline_order=1,
                anti_aliasing=False,
                anti_aliasing_smoothing_beta=0.98
            )

        # Create an index tracked object
        tracker = InteractivePlot(
            image=self,
            mask=mask,
            slice_id=slice_id
        )

        plt.show()

    @staticmethod
    def get_colour_map():
        return "gist_gray"

    @staticmethod
    def get_default_lowest_intensity():
        return None

    @staticmethod
    def get_default_upper_intensity():
        return None

    @staticmethod
    def get_default_ivh_discretisation_method():
        return "fixed_bin_number"

    @staticmethod
    def get_default_ivh_bin_number():
        return 1000

    @staticmethod
    def get_default_ivh_bin_size():
        return None

    def interpolate(
            self,
            interpolate: None | bool = None,
            new_spacing: None | tuple[float, ...] = None,
            translation: None | float | tuple[float, ...] = (0.0, 0.0, 0.0),
            rotation: None | float = 0.0,
            spline_order: None | int = None,
            anti_aliasing: None | bool = None,
            anti_aliasing_smoothing_beta: None | float = None,
            settings: None | SettingsClass = None
    ):

        if (self.separate_slices is None or interpolate is None or spline_order is None or anti_aliasing is
                None or anti_aliasing_smoothing_beta is None) and settings is None:
            raise ValueError("None of the parameters for interpolation can be set.")

        if interpolate is None:
            interpolate = settings.img_interpolate.interpolate
        if new_spacing is None and settings is not None:
            new_spacing = settings.img_interpolate.new_spacing
        if translation is None:
            if settings is not None:
                translation_x = settings.perturbation.translate_x if settings.perturbation.translate_x is not None else 0.0
                translation_y = settings.perturbation.translate_y if settings.perturbation.translate_y is not None else 0.0
                translation_z = settings.perturbation.translate_z if settings.perturbation.translate_z is not None else 0.0
                translation = tuple([translation_z, translation_y, translation_x])
            else:
                translation = (0.0, 0.0, 0.0)
        elif isinstance(translation, float):
            translation = tuple([translation] * 3)

        if rotation is None and settings is not None:
            rotation = settings.perturbation.rotation_angles[0]
        if spline_order is None:
            spline_order = self.get_interpolation_spline_order(settings=settings)
        if anti_aliasing is None:
            anti_aliasing = settings.img_interpolate.anti_aliasing
        if anti_aliasing_smoothing_beta is None:
            anti_aliasing_smoothing_beta = settings.img_interpolate.smoothing_beta

        # Set spacing
        if new_spacing is None or not interpolate:
            # Use original spacing.
            new_spacing = self.image_spacing

        elif self.separate_slices:
            # Use provided spacing, in 2D. Spacing for interpolation across slices is set to the original spacing in
            # case interpolation is only conducted within the slice.
            new_spacing = list(new_spacing)
            new_spacing[0] = self.image_spacing[0]

        else:
            # Use provided spacing, in 3D
            new_spacing = list(new_spacing)

        # Set translation. Check that translation is specified for every direction.
        translation: list[float, float, float] = list(translation)
        if len(translation) == 1:
            translation *= 3
        for ii in range(len(translation)):
            if translation[ii] is None:
                translation[ii] = 0.0

        if self.separate_slices:
            translation[0] = 0.0

        translation: tuple[float, float, float] = tuple(translation)

        return self._interpolate(
            interpolate=interpolate,
            new_spacing=tuple(new_spacing),
            translation=translation,
            rotation=rotation,
            spline_order=spline_order,
            anti_aliasing=anti_aliasing,
            anti_aliasing_smoothing_beta=anti_aliasing_smoothing_beta
        )

    @staticmethod
    def get_interpolation_spline_order(settings: SettingsClass):
        return settings.img_interpolate.spline_order

    def _interpolate(
            self,
            interpolate: bool,
            new_spacing: tuple[float, ...],
            translation: tuple[float, ...],
            rotation: float,
            spline_order: int,
            anti_aliasing: bool,
            anti_aliasing_smoothing_beta: float
    ):
        """Performs interpolation of the image volume"""
        from mirp._image_processing.anti_aliasing import gaussian_preprocess_filter
        from scipy.ndimage import map_coordinates

        # Skip for missing images.
        if self.is_empty() is None:
            return

        # Translate tuples to np.array
        new_spacing = np.array(new_spacing).astype(float)
        translation = np.array(translation).astype(float)

        # Convert translation to [0.0, 1.0 range].
        translation = translation - np.floor(translation)

        # Set translation
        self.translation = translation

        # Rotation around the z-axis (initial axis in numpy)
        rotation_angle = np.radians(rotation)

        # Set rotation
        self.rotation_angle = rotation

        # Skip if nor interpolation, nor affine transformation are required.
        if not interpolate and np.allclose(translation, 0.0) and np.isclose(rotation_angle, 0.0):
            return

        # Check if pre-processing is required
        if anti_aliasing:
            self.set_voxel_grid(
                voxel_grid=gaussian_preprocess_filter(
                    orig_vox=self.get_voxel_grid(),
                    orig_spacing=np.array(self.image_spacing),
                    sample_spacing=new_spacing,
                    param_beta=anti_aliasing_smoothing_beta,
                    mode="nearest",
                    separate_slices=self.separate_slices
                )
            )

        # Determine dimensions of voxel grid after resampling.
        sample_dim = np.ceil(
            np.multiply(np.array(self.image_dimension), np.array(self.image_spacing) / new_spacing)
        ).astype(int)

        # Set grid spacing (i.e. a fractional spacing in input voxel dimensions)
        grid_spacing = new_spacing / np.array(self.image_spacing)

        # Set grid origin according to IBSI reference manual, and update with the translation.
        grid_origin = translation * grid_spacing + 0.5 * (np.array(self.image_dimension) - 1.0) - 0.5 * (np.array(sample_dim) - 1.0) * grid_spacing

        # Set grid origin in world coordinates.
        world_grid_origin = self.to_world_coordinates(x=grid_origin, trim_result=False)

        # Establish the rotation center.
        world_rotation_center = self.to_world_coordinates(x=grid_origin + 0.5 * (sample_dim - 1) * grid_spacing)

        # Compute affine and inverse affine matrix.
        affine_matrix = self.get_affine_matrix()
        inverse_affine_matrix = self.get_affine_matrix(inverse=True)

        # Create affine matrix for rotating around the rotation center.
        #   W' = T * R * T^-1 * W
        #      = A_geometric * W
        # Here T^-1 moves the center of the world space to the rotation center, R performs the rotation and T^-1
        # moves the center of world space to the origin. Since all matrices are square and have the same dimensions,
        # we will first compute T * R * T^-1, and then multiply this with grid coordinates in world coordinates.

        # Set up the rotation matrix for rotation in the y-x plane.
        rotation_matrix = np.identity(4, dtype=float)
        rotation_matrix[1:3, 1:3] = np.array(
            [[np.cos(rotation_angle), np.sin(rotation_angle)],
             [-np.sin(rotation_angle), np.cos(rotation_angle)]]
        )

        # Create the entire affine matrix to perform the rotation around the rotation center.
        geometric_affine_matrix = np.matmul(
            rotation_matrix, self._get_origin_matrix(origin=world_rotation_center, inverse=True))
        geometric_affine_matrix = np.matmul(
            self._get_origin_matrix(origin=world_rotation_center), geometric_affine_matrix)

        # Multiply all matrices so: A_affine ^-1 * A_geometric * A_affine
        matrix = np.matmul(geometric_affine_matrix, affine_matrix)
        matrix = np.matmul(inverse_affine_matrix, matrix)

        # Generate interpolation grid in voxel space.
        voxel_map_z, voxel_map_y, voxel_map_x = np.mgrid[:sample_dim[0], :sample_dim[1], :sample_dim[2]]
        voxel_map_z = voxel_map_z.flatten() * grid_spacing[0] + grid_origin[0]
        voxel_map_y = voxel_map_y.flatten() * grid_spacing[1] + grid_origin[1]
        voxel_map_x = voxel_map_x.flatten() * grid_spacing[2] + grid_origin[2]

        # Set up voxel map coordinates as a 4 x n matrix.
        voxel_map_coordinates = np.array(
            [voxel_map_z, voxel_map_y, voxel_map_x, np.ones(voxel_map_x.shape, dtype=float)]
        )

        # Free up voxel map variables.
        del voxel_map_x, voxel_map_y, voxel_map_z

        # Rotate and translate the voxel grid.
        voxel_map_coordinates = np.matmul(matrix, voxel_map_coordinates)[0:3, :]

        # Interpolate orig_vox on interpolation grid
        sample_voxel_grid = map_coordinates(
            input=self.get_voxel_grid(),
            coordinates=voxel_map_coordinates,
            order=spline_order,
            mode="nearest"
        )

        # Shape map_vox to the correct dimensions.
        sample_voxel_grid = np.reshape(sample_voxel_grid, sample_dim)

        # Orientation changes through rotation.
        self.image_orientation = np.matmul(rotation_matrix, self._get_orientation_matrix())[0:3, 0:3]

        # Update origin
        self.image_origin = tuple(
            np.squeeze(np.matmul(geometric_affine_matrix, world_grid_origin[:, np.newaxis]), axis=1)[0:3]
        )

        # Update spacing and affine matrix.
        self.image_spacing = tuple(new_spacing)

        # Set interpolation
        self.interpolated = True

        # Set interpolation algorithm
        if spline_order == 0:
            self.interpolation_algorithm = "nnb"
        elif spline_order == 1:
            self.interpolation_algorithm = "lin"
        elif spline_order > 1:
            self.interpolation_algorithm = "si" + str(spline_order)

        # Set voxel grid
        self.set_voxel_grid(voxel_grid=sample_voxel_grid)

    def register(
            self,
            image,
            spline_order: None | int = None,
            anti_aliasing: None | bool = None,
            anti_aliasing_smoothing_beta: None | float = None,
            settings: None | SettingsClass = None,
            mode: str = "nearest"
    ):
        if (spline_order is None or anti_aliasing is None or anti_aliasing is None) and settings is None:
            raise ValueError("None of the parameters for registration can be set.")

        if spline_order is None:
            spline_order = self.get_interpolation_spline_order(settings=settings)
        if anti_aliasing is None:
            anti_aliasing = settings.img_interpolate.anti_aliasing
        if anti_aliasing_smoothing_beta is None:
            anti_aliasing_smoothing_beta = settings.img_interpolate.smoothing_beta

        return self._register(
            image=image,
            spline_order=spline_order,
            anti_aliasing=anti_aliasing,
            anti_aliasing_smoothing_beta=anti_aliasing_smoothing_beta,
            mode=mode
        )

    def _register(
            self,
            image,
            spline_order: int,
            anti_aliasing: bool,
            anti_aliasing_smoothing_beta: float,
            mode: str = "nearest"
    ):
        """Register this image with another image."""

        from scipy.ndimage import map_coordinates
        from mirp._image_processing.anti_aliasing import gaussian_preprocess_filter

        # This is just for type hinting. Use typing.Self once this is supported by the codestack.
        image: GenericImage = image

        # Skip if either internal or external image data are missing.
        if self.is_empty() or image.is_empty():
            return

        # Check whether registration is required
        registration_required = False

        # Mismatch in grid dimension
        if not np.array_equal(self.image_dimension, image.image_dimension):
            registration_required = True

        # Mismatch in origin
        if not np.allclose(self.image_origin, image.image_origin):
            registration_required = True

        # Mismatch in spacing
        if not np.allclose(self.image_spacing, image.image_spacing):
            registration_required = True

        # Mismatch in orientation
        if not np.allclose(self.image_orientation, image.image_orientation):
            registration_required = True

        if not registration_required:
            return

        # Apply anti-aliasing.
        if anti_aliasing:
            self.set_voxel_grid(
                voxel_grid=gaussian_preprocess_filter(
                    orig_vox=self.get_voxel_grid(),
                    orig_spacing=np.array(self.image_spacing),
                    sample_spacing=np.array(image.image_spacing),
                    param_beta=anti_aliasing_smoothing_beta,
                    mode="nearest",
                    separate_slices=False
                )
            )

        # Create grid coordinates in world space using the image object.
        grid_coordinates = image.world_coordinates()

        # Translate grid coordinates into voxel space of the current image.
        grid_coordinates = self.to_voxel_coordinates(x=grid_coordinates)

        # Interpolate at the grid coordinates.
        new_mask = map_coordinates(
            input=self.get_voxel_grid().astype(float),
            coordinates=grid_coordinates,
            order=spline_order,
            mode=mode
        )

        # Restore form.
        new_mask = np.reshape(new_mask, image.image_dimension)

        # Update positional and affine parameters of the image.
        self.image_orientation = copy.deepcopy(image.image_orientation)
        self.image_origin = copy.deepcopy(image.image_origin)
        self.image_spacing = copy.deepcopy(image.image_spacing)

        # Update translation and rotation using the image object.
        self.translation = image.translation
        self.rotation_angle = image.rotation_angle

        # Set and update image after registration.
        self.set_voxel_grid(voxel_grid=new_mask)

    def add_noise(self, noise_level, noise_iteration_id):
        """
         Adds Gaussian noise to the image volume
         noise_level: standard deviation of image noise present """

        # Add noise iteration number
        self.noise_iteration_id = noise_iteration_id

        # Skip for missing images
        if self.image_data is None:
            return

        # Skip for invalid noise levels
        if noise_level is None:
            return
        if np.isnan(noise_level) or noise_level < 0.0:
            return

        # Set noise level in image
        self.noise_level = noise_level

        # Start random generator..
        m = hashlib.sha1(usedforsecurity=False)
        m = self.update_hash(m=m)
        randomiser = np.random.default_rng(int(m.hexdigest(), 16))

        # Add Gaussian noise to image
        voxel_grid = self.get_voxel_grid()
        voxel_grid += randomiser.normal(loc=0.0, scale=noise_level, size=self.image_dimension)

        # Update image.
        self.set_voxel_grid(voxel_grid=voxel_grid)

    def estimate_noise(self, method="chang"):

        import scipy.ndimage as ndi

        # Skip if the image is missing
        if self.is_empty():
            return

        if method == "rank":
            """ Estimate image noise level using the method by Rank, Lendl and Unbehauen, Estimation of 
            image noise variance, IEEE Proc. Vis. Image Signal Process (1999) 146:80-84"""

            # Step 1: filter with a cascading difference filter to suppress original image
            difference_filter = np.array([-1.0 / np.sqrt(2.0), 1.0 / np.sqrt(2.0)])

            # Filter voxel volume
            response_map = ndi.convolve1d(self.get_voxel_grid(), weights=difference_filter, axis=1)
            response_map = ndi.convolve1d(response_map, weights=difference_filter, axis=2)

            # Step 2: compute histogram of local standard deviation and calculate histogram
            # Calculate local means
            local_means = ndi.uniform_filter(response_map, size=[1, 3, 3])

            # Calculate local sum of squares
            sum_filter = np.array([1.0, 1.0, 1.0]) / 3.0
            local_sum_square = ndi.convolve1d(np.power(response_map, 2.0), weights=sum_filter, axis=1)
            local_sum_square = ndi.convolve1d(local_sum_square, weights=sum_filter, axis=2)

            # Calculate local variance
            local_variance = 1.0 / 8.0 * (local_sum_square - 9.0 * np.power(local_means, 2.0))

            # Step 3: calculate median noise - this differs from the original
            # Set local variances below 0 (due to floating point rounding) to 0.
            local_variance = np.ravel(local_variance)
            local_variance[local_variance < 0.0] = 0.0

            # Select robust range (within IQR)
            local_variance = local_variance[np.logical_and(
                local_variance >= np.percentile(local_variance, 25),
                local_variance <= np.percentile(local_variance, 75)
            )]

            # Calculate Gaussian noise
            estimated_noise = np.sqrt(np.mean(local_variance))

        elif method == "ikeda":
            """ Estimate image noise level using a method by Ikeda, Makino, Imai et al., A method for estimating noise
             variance of CT image, Comp Med Imaging Graph (2010) 34:642-650"""

            # Step 1: filter with a cascading difference filter to suppress original image volume
            diff_filter = np.array([-1.0 / np.sqrt(2.0), 1.0 / np.sqrt(2.0)])

            # Filter voxel volume
            response_map = ndi.convolve1d(self.get_voxel_grid(), weights=diff_filter, axis=1)
            response_map = ndi.convolve1d(response_map, weights=diff_filter, axis=2)

            # Step 2: calculate median noise
            estimated_noise = np.median(np.abs(response_map)) / 0.6754

        elif method == "chang":
            """ Noise estimation based on wavelets used in Chang, Yu and Vetterli, Adaptive wavelet thresholding for image
            denoising and compression. IEEE Trans Image Proc (2000) 9:1532-1546"""

            import pywt

            # Step 1: calculate HH subband of the wavelet transformation
            # Generate digital wavelet filter
            hi_filt = np.array(pywt.Wavelet("coif1").dec_hi)

            # Calculate HH subband image
            response_map = ndi.convolve1d(self.get_voxel_grid(), weights=hi_filt, axis=1)
            response_map = ndi.convolve1d(response_map, weights=hi_filt, axis=2)

            # Step 2: calculate median noise
            estimated_noise = np.median(np.abs(response_map)) / 0.6754

        elif method == "immerkaer":
            """ Noise estimation based on laplacian filtering, described in Immerkaer, Fast noise variance estimation.
            Comput Vis Image Underst (1995) 64:300-302"""

            # Step 1: construct filter and filter voxel volume
            # Create filter
            noise_filt = np.array([[1.0, -2.0, 1.0], [-2.0, 4.0, -2.0], [1.0, -2.0, 1.0]], ndmin=3)

            # Apply filter
            response_map = ndi.convolve(self.get_voxel_grid(), weights=noise_filt)

            # Step 2: calculate noise level
            estimated_noise = np.sqrt(np.mean(np.power(response_map, 2.0))) / 36.0

        elif method == "zwanenburg":
            """ Noise estimation based on blob detection for weighting immerkaer filtering """

            # Step 1: construct laplacian filter and filter voxel volume
            # Create filter
            noise_filt = np.array([[1.0, -2.0, 1.0], [-2.0, 4.0, -2.0], [1.0, -2.0, 1.0]], ndmin=3)

            # Apply filter
            response_map = ndi.convolve(self.get_voxel_grid(), weights=noise_filt)
            response_map = np.power(response_map, 2.0)

            # Step 2: construct blob weighting
            # Spacing for gaussian
            gauss_filt_spacing = np.full(shape=3, fill_value=np.min(self.image_spacing))
            gauss_filt_spacing = np.divide(gauss_filt_spacing, np.array(self.image_spacing))

            # Difference of gaussians
            weight_vox = ndi.gaussian_filter(
                self.get_voxel_grid(),
                sigma=1.0 * gauss_filt_spacing
            ) - ndi.gaussian_filter(
                self.get_voxel_grid(),
                sigma=4.0 * gauss_filt_spacing
            )

            # Smooth edge detection
            weight_vox = ndi.gaussian_filter(np.abs(weight_vox), sigma=2.0 * gauss_filt_spacing)

            # Convert to weighting scale
            weight_vox = 1.0 - weight_vox / np.max(weight_vox)

            # Decrease weight of edge voxels
            weight_vox = np.power(weight_vox, 2.0)

            # Step 3: estimate noise level
            estimated_noise = np.sqrt(np.sum(np.multiply(response_map, weight_vox)) / (36.0 * np.sum(weight_vox)))

        else:
            raise ValueError(
                "The provided noise estimation method is not implemented. Use one of \"chang\" (default), "
                "\"rank\", \"ikeda\", \"immerkaer\" or \"zwanenburg\"."
            )

        return estimated_noise

    def saturate(self, intensity_range, fill_value=None):
        """
        Saturate image intensities using an intensity range
        :param intensity_range: range of intensity values
        :param fill_value: fill value for out-of-range intensities. If None, the upper and lower ranges are used
        :return:
        """
        # Skip for missing images
        if self.image_data is None or intensity_range is None:
            return

        intensity_range = np.array(copy.deepcopy(intensity_range))

        if np.any(~np.isnan(intensity_range)):
            # Set updated image data.
            image_data = self.get_voxel_grid()

            # Saturate values below lower boundary.
            if not np.isnan(intensity_range[0]):
                if fill_value is None:
                    image_data[image_data < intensity_range[0]] = intensity_range[0]
                else:
                    image_data[image_data < intensity_range[0]] = fill_value[0]

            # Saturate values above upper boundary.
            if not np.isnan(intensity_range[1]):
                if fill_value is None:
                    image_data[image_data > intensity_range[1]] = intensity_range[1]
                else:
                    image_data[image_data > intensity_range[1]] = fill_value[1]

            # Set the updated image data.
            self.set_voxel_grid(voxel_grid=image_data)

    def normalise_intensities(
            self,
            normalisation_method: None | str = "none",
            intensity_range: None | tuple[Any, Any] = None,
            saturation_range: None | tuple[Any, Any] = None,
            mask: None | np.ndarray = None
    ) -> Self:
        """
        Normalises image intensities
        :param normalisation_method: string defining the normalisation method. Should be one of "none", "range",
        "relative_range", "quantile_range", "standardisation".
        :param intensity_range: range of intensities for normalisation.
        :param saturation_range: range of allowed intensity values.
        :param mask: sets area that should be considered for determining normalisation parameters.
        :return:
        """

        # Skip for missing images
        if self.is_empty():
            return self

        if intensity_range is None:
            intensity_range = [np.nan, np.nan]

        if mask is None:
            mask = np.ones(self.image_dimension, dtype=bool)
        else:
            mask = mask.astype(bool)

        if np.sum(mask) == 0:
            mask = np.ones(self.image_dimension, dtype=bool)

        if saturation_range is None:
            saturation_range = [np.nan, np.nan]

        if normalisation_method is None or normalisation_method == "none":
            return self

        elif normalisation_method == "range":
            # Normalisation to [0, 1] range using fixed intensities.

            # Get image data
            image_data = self.get_voxel_grid()

            # Find maximum and minimum intensities
            if np.isnan(intensity_range[0]):
                min_int = np.min(image_data[mask])
            else:
                min_int = intensity_range[0]

            if np.isnan(intensity_range[1]):
                max_int = np.max(image_data[mask])
            else:
                max_int = intensity_range[1]

            # Normalise by range
            if not max_int == min_int:
                image_data = (image_data - min_int) / (max_int - min_int)
            else:
                image_data = image_data - min_int

            # Update image data
            self.set_voxel_grid(voxel_grid=image_data)

        elif normalisation_method == "relative_range":
            # Normalisation to [0, 1]-ish range using relative intensities.

            # Get image data
            image_data = self.get_voxel_grid()

            min_int_rel = 0.0
            if not np.isnan(intensity_range[0]):
                min_int_rel = intensity_range[0]

            max_int_rel = 1.0
            if not np.isnan(intensity_range[1]):
                max_int_rel = intensity_range[1]

            # Compute minimum and maximum intensities.
            value_range = [np.min(image_data[mask]), np.max(image_data[mask])]
            min_int = value_range[0] + min_int_rel * (value_range[1] - value_range[0])
            max_int = value_range[0] + max_int_rel * (value_range[1] - value_range[0])

            # Normalise by range
            if not max_int == min_int:
                image_data = (image_data - min_int) / (max_int - min_int)
            else:
                image_data = image_data - min_int

            # Update image data
            self.set_voxel_grid(voxel_grid=image_data)

        elif normalisation_method == "quantile_range":
            # Normalisation to [0, 1]-ish range based on quantiles.

            # Get image data
            image_data = self.get_voxel_grid()

            min_quantile = 0.0
            if not np.isnan(intensity_range[0]):
                min_quantile = intensity_range[0]

            max_quantile = 1.0
            if not np.isnan(intensity_range[1]):
                max_quantile = intensity_range[1]

            # Compute quantiles from voxel grid.
            min_int = np.quantile(image_data[mask], q=min_quantile)
            max_int = np.quantile(image_data[mask], q=max_quantile)

            # Normalise by range
            if not max_int == min_int:
                image_data = (image_data - min_int) / (max_int - min_int)
            else:
                image_data -= min_int

            # Update image data
            self.set_voxel_grid(voxel_grid=image_data)

        elif normalisation_method == "standardisation":
            # Normalisation to mean 0 and standard deviation 1.

            # Get image data
            image_data = self.get_voxel_grid()

            # Determine mean and standard deviation of the voxel intensities
            mean_int = np.mean(image_data[mask])
            sd_int = np.std(image_data[mask])

            # Protect against invariance.
            if sd_int == 0.0:
                sd_int = 1.0

            # Normalise
            image_data = (image_data - mean_int) / sd_int

            # Update image data
            self.set_voxel_grid(voxel_grid=image_data)
        else:
            raise ValueError(f"{normalisation_method} is not a valid method for normalising intensity values.")

        self.saturate(intensity_range=saturation_range)

        return self

    def scale_intensities(self, scale: float) -> Self:

        # Skip for missing images
        if self.is_empty():
            return self

        if scale == 1.0:
            return self
        elif scale == 0.0:
            raise ValueError("scale cannot be 0.0, as this would result in an image with 0.0 values only.")

        # Get image data.
        image_data = self.get_voxel_grid()
        image_data *= scale

        self.set_voxel_grid(image_data)

        return self

    def decimate(self):
        """
        Decimates image by removing every second element
        """

        # Skip for missing images
        if self.is_empty():
            return

        # Get the voxel grid
        image_data = self.get_voxel_grid()
        image_spacing = np.array(self.image_spacing)

        # Update the voxel grid
        if self.separate_slices:
            # Drop every second pixel
            image_data = image_data[:, slice(None, None, 2), slice(None, None, 2)]
            self.image_spacing = tuple(image_spacing[[1, 2]] * 2.0)

        else:
            # Drop every second voxel
            image_data = image_data[slice(None, None, 2), slice(None, None, 2), slice(None, None, 2)]

            # Update voxel spacing
            self.image_spacing = tuple(image_spacing * 2.0)

        # Update voxel grid. This also updates the size attribute.
        self.set_voxel_grid(voxel_grid=image_data)

    def crop(
            self,
            ind_ext_z=None,
            ind_ext_y=None,
            ind_ext_x=None,
            xy_only=False,
            z_only=False
    ):
        """Crop image to the provided map extent."""

        # Skip for missing images
        if self.is_empty():
            return

        # Determine corresponding voxel indices
        max_ind = np.ceil(np.array((np.max(ind_ext_z), np.max(ind_ext_y), np.max(ind_ext_x)))).astype(int)
        min_ind = np.floor(np.array((np.min(ind_ext_z), np.min(ind_ext_y), np.min(ind_ext_x)))).astype(int)

        # Set bounding indices
        max_bound_ind = np.minimum(max_ind, np.array(self.image_dimension)).astype(int)
        min_bound_ind = np.maximum(min_ind, np.array([0, 0, 0])).astype(int)

        # Get image data.
        image_data = self.get_voxel_grid()

        # Create corresponding image volumes by slicing original volume
        if z_only:
            image_data = image_data[
                min_bound_ind[0]:max_bound_ind[0] + 1,
                :,
                :
            ]
            min_bound_ind[1] = 0
            min_bound_ind[2] = 0
        elif xy_only:
            image_data = image_data[
                :,
                min_bound_ind[1]:max_bound_ind[1] + 1,
                min_bound_ind[2]:max_bound_ind[2] + 1
            ]
            min_bound_ind[0] = 0
            max_bound_ind[0] = np.array(self.image_dimension)[0].astype(int)
        else:
            image_data = image_data[
                min_bound_ind[0]:max_bound_ind[0] + 1,
                min_bound_ind[1]:max_bound_ind[1] + 1,
                min_bound_ind[2]:max_bound_ind[2] + 1
            ]

        # Update origin
        self.image_origin = tuple(self.to_world_coordinates(x=np.array(min_bound_ind)))

        # Update voxel grid
        self.set_voxel_grid(voxel_grid=image_data)

    def crop_to_size(self, center, crop_size):
        """Crop images to the exact size"""

        # Skip for missing images
        if self.is_empty():
            return

        # Make local copies.
        crop_size = copy.deepcopy(crop_size)
        center = copy.deepcopy(center)

        if len(crop_size) == 2:
            crop_size = [None, crop_size[0], crop_size[1]]
        if len(center) == 2:
            center = [None, center[0], center[1]]

        # Set origin for the new grid.
        new_grid_index = [
            int(0.5 * (center[ii] - crop_size[ii]))
            if center[ii] is not None and crop_size[ii] is not None else 0
            for ii in range(len(center))
        ]

        # Update crop size.
        crop_size = [
            crop_size[ii]
            if crop_size[ii] is not None else self.image_dimension[ii]
            for ii in range(len(crop_size))
        ]
        crop_size = np.array(crop_size)
        new_grid_index = np.array(new_grid_index)

        # Determine coordinates of the box that can be copied in the original space
        max_index_original = new_grid_index + crop_size
        min_index_original = new_grid_index

        # Update coordinates based on boundaries in the original images
        max_index_original = np.minimum(max_index_original, np.array(self.image_dimension))
        min_index_original = np.maximum(min_index_original, np.array([0, 0, 0]))

        # Determine coordinates where this box should land, i.e. perform the coordinate transformation to grid index space.
        max_index_crop = max_index_original - new_grid_index
        min_index_crop = min_index_original - new_grid_index

        # Create an empty voxel_grid to copy to
        cropped_image = np.full(crop_size, fill_value=np.nan)

        # Put the voxel grid slice into the cropped grid
        cropped_image[
            min_index_crop[0]:max_index_crop[0],
            min_index_crop[1]:max_index_crop[1],
            min_index_crop[2]:max_index_crop[2]
        ] = self.get_voxel_grid()[
            min_index_original[0]:max_index_original[0],
            min_index_original[1]:max_index_original[1],
            min_index_original[2]:max_index_original[2]
        ]

        # Replace any remaining NaN values in the grid by the lowest intensity in voxel_grid
        cropped_image[np.isnan(cropped_image)] = np.min(self.get_voxel_grid())

        # Restore the original dtype in case it got lost
        cropped_image = cropped_image.astype(self.get_voxel_grid().dtype)

        # Update origin
        self.image_origin = tuple(self.to_world_coordinates(x=np.array(new_grid_index)))

        # Set voxel grid
        self.set_voxel_grid(voxel_grid=cropped_image)

    def get_supervoxels(
            self,
            intensity_range: tuple[float, float]
    ):
        """Extracts supervoxels from an image"""

        from skimage.segmentation import slic
        from mirp._image_processing.utilities import extend_intensity_range
        from mirp._image_processing.utilities import set_intensity_range

        if self.is_empty():
            return None

        # Update or set intensity range, and extend it by around 10% on either side.
        intensity_range = set_intensity_range(image=self, intensity_range=intensity_range)
        intensity_range = extend_intensity_range(intensity_range=intensity_range, extend_fraction=0.1)

        # Get image data
        image_data = copy.deepcopy(self.get_voxel_grid())

        # Apply threshold
        image_data[image_data < intensity_range[0]] = intensity_range[0]
        image_data[image_data > intensity_range[1]] = intensity_range[1]

        # Slic constants - sigma
        sigma = 1.0 * np.min(self.image_spacing)

        # Slic constants - number of segments.
        min_n_voxels = np.max([20.0, 500.0 / np.prod(self.image_spacing)])
        n_segments = int(np.prod(self.image_dimension) / min_n_voxels)

        # Convert to float with range [0.0, 1.0]
        image_data -= intensity_range[0]
        image_data *= 1.0 / (intensity_range[1] - intensity_range[0])

        if image_data.dtype not in ["float", "float64"]:
            image_data = image_data.astype(float)

        # Create a slic segmentation of the image stack
        image_segments = slic(
            image=image_data,
            n_segments=n_segments,
            sigma=sigma,
            spacing=self.image_spacing,
            compactness=0.05,
            convert2lab=False,
            enforce_connectivity=True,
            channel_axis=None
        )

        image_segments += 1

        # Release image_data
        del image_data

        return image_segments

    def bias_field_correction(self, in_place=True, **kwargs):
        if not in_place:
            return self.copy()

    def write(
            self,
            dir_path: str | Path,
            file_name: None | str = None,
            file_format: str = "nifti"
    ):
        """ Writes the image to a file """
        import os
        import itk

        if self.is_empty():
            return

        # Check if path exists
        dir_path = os.path.normpath(dir_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        # Generate filename, if necessary.
        if file_name is None:
            file_name = "_".join(self.get_file_name_descriptor())

        # Add extension.
        if file_format == "nifti":
            file_name += ".nii.gz"
        elif file_format == "numpy":
            file_name += ".npy"
        else:
            raise ValueError(f"The provided file format {file_format} is not available.")

        # Add file and file name
        file_path = os.path.join(dir_path, file_name)

        if file_format == "nifti":
            image_data = self.get_voxel_grid()
            if np.issubdtype(image_data.dtype, bool):
                cast_type = np.uint8
            elif np.issubdtype(image_data.dtype, float):
                cast_type = float
            else:
                cast_type = image_data.dtype

            image_data = itk.GetImageFromArray(image_data.astype(cast_type))
            image_data.SetOrigin(np.array(self.image_origin)[::-1])
            image_data.SetSpacing(np.array(self.image_spacing)[::-1])
            image_data.SetDirection(itk.matrix_from_array(np.reshape(np.ravel(self.image_orientation)[::-1], [3, 3])))

            itk.imwrite(image_data, file_path)

        elif file_format == "numpy":
            image_data = self.get_voxel_grid()
            np.save(file_path, image_data)

        else:
            raise ValueError(
                f"file format was not recognised: {file_format}. MIRP currently supports nifti and numpy are supported."
            )

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = []

        # Sample name
        if self.sample_name is not None:
            descriptors += [self.sample_name]

        # Slice id
        if self.slice_id is not None:
            descriptors += [self.slice_id]

        # Modality
        if self.modality is not None:
            descriptors += [self.modality]

        # Interpolation
        if self.interpolated:
            descriptors += [
                self.interpolation_algorithm,
                "x", str(self.image_spacing[2])[:5],
                "y", str(self.image_spacing[1])[:5],
                "z", str(self.image_spacing[0])[:5]
            ]

        # Rotation
        if self.rotation_angle is not None and self.rotation_angle != 0.0:
            descriptors += ["rot", str(self.rotation_angle)[:5]]

        # Translation
        if self.translation is not None and not np.all(np.array(self.translation) == 0.0):
            descriptors += [
                "trans",
                "x", str(self.translation[2])[:5],
                "y", str(self.translation[1])[:5],
                "z", str(self.translation[0])[:5]
            ]

        # Noise
        if self.noise_level is not None and self.noise_level > 0.0:
            descriptors += ["noise", str(self.noise_level)[:5], "id", str(self.noise_iteration_id)]

        return descriptors

    def export(
            self,
            export_format: str = "dict"
    ) -> None | np.ndarray | dict[str, Any] | Self:

        if self.is_empty():
            return None

        if export_format == "dict":
            attributes = self.get_export_attributes()
            attributes.update({"image": self.get_voxel_grid()})

            return attributes
        elif export_format == "numpy":
            return self.get_voxel_grid()

        elif export_format == "native":
            return self.copy()

        else:
            raise ValueError(f"The current value of export_format was not recognised: {export_format}")

    def export_metadata(self, **kwargs) -> None | dict[str, Any]:
        metadata = self.object_metadata
        if isinstance(metadata, dict):
            return metadata
        else:
            return dict()

    def get_export_attributes(self) -> dict[str, Any]:
        attributes = []

        # Sample name
        if self.sample_name is not None:
            attributes += [("sample_name", self.sample_name)]

        # Slice id.
        if self.slice_id is not None:
            attributes += [("slice_id", self.slice_id)]

        # Modality
        if self.modality is not None:
            attributes += [("modality", self.modality)]

        # Rotation
        if self.rotation_angle is None:
            attributes += [("rotation", 0.0)]
        else:
            attributes += [("rotation", self.rotation_angle)]

        # Translation
        if self.translation is None:
            attributes += [("translation", (0.0, 0.0, 0.0))]
        else:
            attributes += [("translation", self.translation)]

        # Noise
        if self.noise_level is not None and self.noise_level > 0.0:
            attributes += [("noise_level", self.noise_level), ("noise_id", self.noise_iteration_id)]

        # Image spacing, origin and orientation.
        attributes += [
            ("image_spacing", self.image_spacing),
            ("image_origin", self.image_origin),
            ("image_orientation", self.image_orientation)
        ]

        attributes = dict(attributes)
        if len(self.object_metadata) > 0:
            attributes.update(self.object_metadata)

        return dict(attributes)

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        feature_name_suffix = []
        if self.discretisation_method is not None:
            if self.discretisation_method == "none":
                pass
            elif self.discretisation_method == "fixed_bin_number":
                feature_name_suffix += ["fbn"]
                feature_name_suffix += ["n" + str(self.discretisation_bin_number)]
            elif self.discretisation_method == "fixed_bin_size":
                feature_name_suffix += ["fbs"]
                feature_name_suffix += ["w" + str(self.discretisation_bin_width)]
            elif self.discretisation_method == "fixed_bin_size_pyradiomics":
                feature_name_suffix += ["fbsp"]
                feature_name_suffix += ["w" + str(self.discretisation_bin_width)]
            else:
                raise ValueError(f"Unknown discretisation method: {self.discretisation_method}")

        if len(feature_name_suffix) > 0:
            feature_name_suffix = "_".join(feature_name_suffix)
            feature_name_suffix = "_" + feature_name_suffix
            x.columns += feature_name_suffix

        return x

    def update_hash(self, m):
        # Hash the image object itself.
        m.update(self.get_voxel_grid().tobytes())

        # Hash the image attributes.
        image_attributes = self.get_export_attributes()
        m.update(str(image_attributes).encode())

        return m



import copy
import itertools
import warnings

import numpy as np
import pandas as pd

from mirp._data_import.generic_file import ImageFile, MaskFile
from mirp._data_import.dicom_file import ImageDicomFile, MaskDicomFile
from mirp._data_import.itk_file import ImageITKFile, MaskITKFile
from mirp._data_import.numpy_file import ImageNumpyFile, MaskNumpyFile


class ImageFileStack(ImageFile):
    def is_stackable(self, stack_images: str):
        return False

    def __init__(
            self,
            image_file_objects: list[ImageFile] | list[ImageDicomFile] | list[ImageITKFile] | list[ImageNumpyFile],
            dir_path: None | str = None,
            sample_name: None | str = None,
            image_name: None | str | list[str] = None,
            image_modality: None | str = None,
            image_file_type: None | str = None,
            **kwargs
    ):

        if dir_path is None:
            dir_path = image_file_objects[0].dir_path

        if sample_name is None:
            sample_name = image_file_objects[0].sample_name

        if image_name is None:
            image_name = image_file_objects[0].image_name

        if image_modality is None:
            image_modality = image_file_objects[0].modality

        if image_file_type is None:
            image_file_type = image_file_objects[0].file_type

        if len(image_file_objects) == 1:
            raise ValueError(f"DEV: More than one file is expected for file stacks.")

        # Aspects regarding the image itself are set based on the stack itself.
        super().__init__(
            file_path=None,
            dir_path=dir_path,
            sample_name=sample_name,
            file_name=None,
            image_name=image_name,
            image_modality=image_modality,
            image_file_type=image_file_type,
            image_data=None,
            image_origin=None,
            image_orientation=None,
            image_spacing=None,
            image_dimensions=None,
            **kwargs
        )

        self.image_file_objects = image_file_objects
        self.slice_positions: None | list[float] = None

    def create(self):
        # Import locally to avoid potential circular references.
        from mirp._data_import.dicom_file_stack import ImageDicomFileStack
        from mirp._data_import.itk_file_stack import ImageITKFileStack
        from mirp._data_import.numpy_file_stack import ImageNumpyFileStack

        if all(isinstance(image_file_object, ImageDicomFile) for image_file_object in self.image_file_objects):
            file_stack_class = ImageDicomFileStack
            file_type = "dicom"

        elif all(isinstance(image_file_object, ImageITKFile) for image_file_object in self.image_file_objects):
            file_stack_class = ImageITKFileStack
            file_type = self.image_file_objects[0].file_type

        elif all(isinstance(image_file_object, ImageNumpyFile) for image_file_object in self.image_file_objects):
            file_stack_class = ImageNumpyFileStack
            file_type = "numpy"

        else:
            raise TypeError(
                f"The list of image objects does not consist of a known object type. [{self.describe_self()}]"
            )

        image_file_stack = file_stack_class(
            image_file_objects=self.image_file_objects,
            dir_path=self.dir_path,
            sample_name=self.sample_name,
            image_name=self.image_name,
            image_modality=self.modality,
            image_file_type=file_type
        )

        return image_file_stack

    def complete(self, remove_metadata=False, force=False):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of complete. Please specify "
            f"implementation for subclasses."
        )

    def _complete_sample_name(self):
        if self.sample_name is None:
            self.image_file_objects[0]._complete_sample_name()
            self.sample_name = self.image_file_objects[0].sample_name

    def _complete_modality(self):
        if self.modality is None:
            self.image_file_objects[0]._complete_modality()
            self.modality = self.image_file_objects[0].modality

    def _complete_image_origin(self, force=False, frame_id=None):
        # Image origin and other image-related aspects are set using the complete method of subclasses.
        pass

    def _complete_image_orientation(self, force=False, frame_id=None):
        pass

    def _complete_image_spacing(self, force=False, frame_id=None):
        pass

    def _complete_image_dimensions(self, force=False):
        pass

    def sort_image_objects_by_file(self):
        """
        Strip sample name and any image name from filenames. Then isolate numeric values.
        sequences of numeric values. We follow the following rules:
        1. Check if all files have a numeric value in their name, otherwise, use the original order.
        2. Check that all files only have a single range of numeric values (otherwise, it might hard to arrange and
        identify sequences).
        3. Sort and check that sequences are truly sequential, i.e. have a difference of one.
        :return: nothing, changes are made in-place.
        """

        file_name_numeric = [image_object._get_numeric_sequence_from_file() for image_object in self.image_file_objects]
        if any(current_file_name_numeric is None for current_file_name_numeric in file_name_numeric):
            warnings.warn(
                f"Cannot form stacks from numpy slices based on the file name as numeric values are missing "
                f"from one or more files. The original file order is used. [{self.describe_self()}]", UserWarning
            )
            return

        if any(len(current_file_name_numeric) > 1 for current_file_name_numeric in file_name_numeric):
            warnings.warn(
                f"Cannot form stacks from numpy slices based on the file name as more than one sequence of numeric "
                f"values are present in the name of one or more files. This excludes the sample name (if known) and "
                f"any identifiers for image data. The original file order is used. [{self.describe_self()}]",
                UserWarning
            )
            return

        # Flatten array and convert to integer values.
        file_name_numeric = list(itertools.chain.from_iterable(file_name_numeric))
        file_name_numeric = [int(current_file_name_numeric) for current_file_name_numeric in file_name_numeric]

        if len(file_name_numeric) == 1:
            return

        # Check that all numbers are sequential.
        if not np.all(np.diff(np.sort(np.array(file_name_numeric))) == 1):
            warnings.warn(
                f"Cannot form stacks from numpy slices based on the file name as numbers are not fully sequential for"
                f" all files. The original file order is used. [{self.describe_self()}]",
                UserWarning
            )
            return

        position_table = pd.DataFrame({
            "original_object_order": list(range(len(self.image_file_objects))),
            "order_id": file_name_numeric,
        }).sort_values(by=["order_id"])

        # Sort image file objects.
        self.image_file_objects = [
            self.image_file_objects[position_table.original_object_order[ii]]
            for ii in range(len(position_table))
        ]

    def load_metadata(self, limited=False, include_image=False):
        # Load metadata for underlying files in the order indicated by self.image_file_objects.
        for image_file_object in self.image_file_objects:
            image_file_object.load_metadata(limited=limited, include_image=include_image)

    def remove_metadata(self):
        for image_file_object in self.image_file_objects:
            image_file_object.remove_metadata()

    def load_data(self, **kwargs):
        # Load data for underlying files in the order indicated by self.image_file_objects.
        for image_file_object in self.image_file_objects:
            image_file_object.load_data(**kwargs)

    def stack_slices(self):
        if self.image_data is not None:
            return

        image = np.zeros(self.image_dimension, dtype=np.float32)
        for ii, image_file in enumerate(self.image_file_objects):
            if image_file.image_data is None:
                raise ValueError(
                    "DEV: the image_data attribute of underlying image files are not set. Please call load_data first.")
            image[ii, :, :] = image_file.image_data.astype(np.float32)

        self.image_data = image


class MaskFileStack(ImageFileStack, MaskFile):

    def __init__(
            self,
            image_file_objects: list[MaskFile] | list[MaskDicomFile] | list[MaskITKFile] | list[MaskNumpyFile],
            **kwargs
    ):

        super().__init__(image_file_objects=image_file_objects, **kwargs)

    def complete(self, remove_metadata=False, force=False):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of complete. Please specify "
            f"implementation for subclasses."
        )

    def create(self):
        # Import locally to avoid potential circular references.
        from mirp._data_import.dicom_file_stack import MaskDicomFileStack
        from mirp._data_import.itk_file_stack import MaskITKFileStack
        from mirp._data_import.numpy_file_stack import MaskNumpyFileStack

        if all(isinstance(image_file_object, MaskDicomFile) for image_file_object in self.image_file_objects):
            file_stack_class = MaskDicomFileStack
            file_type = "dicom"

        elif all(isinstance(image_file_object, MaskITKFile) for image_file_object in self.image_file_objects):
            file_stack_class = MaskITKFileStack
            file_type = self.image_file_objects[0].file_type

        elif all(isinstance(image_file_object, MaskNumpyFile) for image_file_object in self.image_file_objects):
            file_stack_class = MaskNumpyFileStack
            file_type = "numpy"

        else:
            raise TypeError(
                f"The list of image objects does not consist of a known object type. [{self.describe_self()}]"
            )

        image_file_stack = file_stack_class(
            image_file_objects=self.image_file_objects,
            dir_path=self.dir_path,
            sample_name=self.sample_name,
            image_name=self.image_name,
            image_modality=self.modality,
            image_file_type=file_type,
            roi_name=self.roi_name
        )

        return image_file_stack

    def stack_slices(self):
        if self.image_data is not None:
            return

        image = np.zeros(self.image_dimension, dtype=int)
        for ii, image_file in enumerate(self.image_file_objects):
            if image_file.image_data is None:
                raise ValueError(
                    "DEV: the image_data attribute of underlying mask files are not set. Please call load_data first."
                )
            image[ii, :, :] = image_file.image_data.astype(int)

        self.image_data = image



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.morph_3d_data import (
    Data3DMesh, Data3DConvexHull, Data3DAxisAlignedBoundingBox, Data3DOrientedMinimumBoundingBox,
    Data3DPrincipleComponents, Data3DSpatial, Data3DMinimumEnvelopingEllipsoid
)
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.base_feature import Feature
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class Feature3DMorph(Feature):

    def __init__(self, allow_approximation: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.allow_approximation = allow_approximation

    def create_table_name(self):
        table_elements = self._get_base_table_name_element()
        self.table_name = "_".join(table_elements)


class Feature3DMesh(Feature3DMorph):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DMesh:
        data = Data3DMesh()
        data.compute(image=image, mask=mask)

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    @staticmethod
    def _compute(data: Data3DMesh, image: GenericImage | None = None, mask: BaseMask | None = None):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = self._compute(data=data, image=image, mask=mask)


class Feature3DConvexHull(Feature3DMesh):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DConvexHull):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DConvexHull:
        # Get parent_data from cache.
        parent_data = Feature3DMesh._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DConvexHull()
        data.__dict__.update(parent_data.__dict__)

        # Compute convex hull vertices.
        data.compute_convex_hull()

        return data

    @staticmethod
    def _compute(
            data: Data3DConvexHull,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DAxisAlignedBoundingBox(Feature3DConvexHull):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DAxisAlignedBoundingBox:
        # Get parent_data from cache.
        parent_data = Feature3DConvexHull._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DAxisAlignedBoundingBox()
        data.__dict__.update(parent_data.__dict__)

        # Compute bounding box volume and area.
        data.compute_bounding_box()

        return data

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DConvexHull):
            Feature3DConvexHull._get_data.cache_clear()
        if not isinstance(other, Feature3DAxisAlignedBoundingBox):
            self._get_data.cache_clear()

    @staticmethod
    def _compute(
            data: Data3DAxisAlignedBoundingBox,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DOrientedMinimumBoundingBox(Feature3DConvexHull):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DOrientedMinimumBoundingBox:
        # Get parent_data from cache.
        parent_data = Feature3DConvexHull._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DOrientedMinimumBoundingBox()
        data.__dict__.update(parent_data.__dict__)

        # Compute bounding box volume and area.
        data.compute_bounding_box()

        return data

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DConvexHull):
            Feature3DConvexHull._get_data.cache_clear()
        if not isinstance(other, Feature3DOrientedMinimumBoundingBox):
            self._get_data.cache_clear()

    @staticmethod
    def _compute(
            data: Data3DOrientedMinimumBoundingBox,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DPCA(Feature3DMesh):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DPrincipleComponents:
        # Get parent_data from cache.
        parent_data = Feature3DMesh._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DPrincipleComponents()
        data.__dict__.update(parent_data.__dict__)

        # Compute semi-axes using principle component analysis.
        data.compute_semi_axes()

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DPCA):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty() or data.is_singular() or data.semi_axes is None:
            self.value = np.nan
        else:
            self.value = self._compute(data=data, image=image, mask=mask)

    @staticmethod
    def _compute(
            data: Data3DPrincipleComponents,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DMinimumEnvelopingEllipsoid(Feature3DConvexHull):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> Data3DMinimumEnvelopingEllipsoid:
        # Get parent_data from cache. Ignore the
        parent_data = Feature3DConvexHull._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DMinimumEnvelopingEllipsoid()
        data.__dict__.update(parent_data.__dict__)

        # Compute semi-axes for the minimum enveloping ellipsoid.
        data.compute_semi_axes()

        return data

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DConvexHull):
            Feature3DConvexHull._get_data.cache_clear()
        if not isinstance(other, Feature3DMinimumEnvelopingEllipsoid):
            self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty() or data.is_singular() or data.semi_axes is None:
            self.value = np.nan
        else:
            self.value = self._compute(data=data, image=image, mask=mask)

    @staticmethod
    def _compute(
            data: Data3DMinimumEnvelopingEllipsoid,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DSpatial(Feature3DMesh):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_spatial_data(
            image: GenericImage,
            mask: BaseMask,
            allow_approximation: bool
    ) -> Data3DSpatial:
        # Get parent_data from cache.
        parent_data = Feature3DMesh._get_data(image=image, mask=mask)

        # Instantiate child using parent attributes.
        data = Data3DSpatial()
        data.__dict__.update(parent_data.__dict__)

        # Compute bounding box volume and area.
        data.compute_spatial_information(
            image=image,
            mask=mask,
            allow_approximation=allow_approximation
        )

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, Feature3DMesh):
            Feature3DMesh._get_data.cache_clear()
        if not isinstance(other, Feature3DSpatial):
            self._get_spatial_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_spatial_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_spatial_data(image=image, mask=mask, allow_approximation=self.allow_approximation)

        # Compute feature value.
        if data.is_empty() or data.is_singular():
            self.value = np.nan
        else:
            self.value = self._compute(data=data, image=image, mask=mask)

    @staticmethod
    def _compute(
            data: Data3DSpatial,
            image: GenericImage | None = None,
            mask: BaseMask | None = None
    ):
        raise NotImplementedError("Implement _compute for feature-specific computation.")


class Feature3DMorphVolume(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - volume"
        self.abbr_name = "morph_volume"
        self.ibsi_id = "RNU0"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        return data.volume


class Feature3DMorphApproximateVolume(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - approximate volume"
        self.abbr_name = "morph_vol_approx"
        self.ibsi_id = "YEKZ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        return len(data.data_morph) * np.prod(data.spacing)


class Feature3DMorphSurfaceArea(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - surface area"
        self.abbr_name = "morph_area_mesh"
        self.ibsi_id = "C0JK"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        return data.area


class Feature3DMorphSurfaceVolumeRatio(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - surface to volume ratio"
        self.abbr_name = "morph_av"
        self.ibsi_id = "2PR5"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        return data.area / data.volume


class Feature3DMorphCompactness1(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - compactness 1"
        self.abbr_name = "morph_comp_1"
        self.ibsi_id = "SKGS"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        x = 36.0 * np.pi * data.volume ** 2.0 / data.area ** 3.0
        return 1.0 / (6.0 * np.pi) * x ** (1.0 / 2.0)


class Feature3DMorphCompactness2(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - compactness 2"
        self.abbr_name = "morph_comp_2"
        self.ibsi_id = "BQWJ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        x = 36.0 * np.pi * data.volume ** 2.0 / data.area ** 3.0
        return x


class Feature3DMorphSphericalDisproportion(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - spherical disproportion"
        self.abbr_name = "morph_sph_dispr"
        self.ibsi_id = "KRCK"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        x = 36.0 * np.pi * data.volume ** 2.0 / data.area ** 3.0
        return x ** (-1.0 / 3.0)


class Feature3DMorphSphericity(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - sphericity"
        self.abbr_name = "morph_sphericity"
        self.ibsi_id = "QCFX"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        x = 36.0 * np.pi * data.volume ** 2.0 / data.area ** 3.0
        return x ** (1.0 / 3.0)


class Feature3DMorphAsphericity(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - asphericity"
        self.abbr_name = "morph_asphericity"
        self.ibsi_id = "25C7"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        x = 36.0 * np.pi * data.volume ** 2.0 / data.area ** 3.0
        return x ** (-1.0 / 3.0) - 1.0


class Feature3DMorphCentreOfMassShift(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - centre of mass shift"
        self.abbr_name = "morph_com"
        self.ibsi_id = "KLMA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        if len(data.data_int) == 0 or len(data.data_morph) == 0:
            return np.nan

        # Compute centre of mass for morphological and intensity-weighted mask.
        com_morph = np.array([
            np.mean(data.data_morph.z),
            np.mean(data.data_morph.y),
            np.mean(data.data_morph.x)
        ])
        com_int = np.array([
            np.sum(data.data_int.g * data.data_int.z),
            np.sum(data.data_int.g * data.data_int.y),
            np.sum(data.data_int.g * data.data_int.x)
        ]) / np.sum(data.data_int.g)

        # Calculate shift in centre of mass.
        return np.sqrt(np.sum(np.multiply(com_morph - com_int, data.spacing) ** 2.0))


class Feature3DMorphIntegratedIntensity(Feature3DMesh):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - "
        self.abbr_name = "morph_integ_int"
        self.ibsi_id = "99N0"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DMesh, **kwargs) -> float:
        return data.volume * np.mean(data.data_int.g)


class Feature3DMorphMaximum3DDiameter(Feature3DConvexHull):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - maximum 3D diameter"
        self.abbr_name = "morph_diam"
        self.ibsi_id = "L0JK"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DConvexHull, **kwargs) -> float:
        from scipy.spatial.distance import pdist
        return np.max(pdist(data.convex_hull_vertices))


class Feature3DMorphConvexHullVolumeDensity(Feature3DConvexHull):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - convex hull volume density"
        self.abbr_name = "morph_vol_dens_conv_hull"
        self.ibsi_id = "R3ER"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DConvexHull, **kwargs) -> float:
        return data.volume / data.convex_hull_volume


class Feature3DMorphConvexHullAreaDensity(Feature3DConvexHull):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - convex hull area density"
        self.abbr_name = "morph_area_dens_conv_hull"
        self.ibsi_id = "7T7F"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DConvexHull, **kwargs) -> float:
        return data.area / data.convex_hull_area


class Feature3DMorphAxisAlignedBoundingBoxVolumeDensity(Feature3DAxisAlignedBoundingBox):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - axis-aligned bounding box volume density"
        self.abbr_name = "morph_vol_dens_aabb"
        self.ibsi_id = "PBX1"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DAxisAlignedBoundingBox, **kwargs) -> float:
        return data.volume / data.bounding_box_volume


class Feature3DMorphAxisAlignedBoundingBoxAreaDensity(Feature3DAxisAlignedBoundingBox):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - axis-aligned bounding box area density"
        self.abbr_name = "morph_area_dens_aabb"
        self.ibsi_id = "R59B"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DAxisAlignedBoundingBox, **kwargs) -> float:
        return data.area / data.bounding_box_area


class Feature3DMorphOrientedMinimumBoundingBoxVolumeDensity(Feature3DOrientedMinimumBoundingBox):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - oriented minimum bounding box volume density"
        self.abbr_name = "morph_vol_dens_ombb"
        self.ibsi_id = "ZH1A"
        self.ibsi_compliant = False

    @staticmethod
    def _compute(data: Data3DOrientedMinimumBoundingBox, **kwargs) -> float:
        return data.volume / data.bounding_box_volume


class Feature3DMorphOrientedMinimumBoundingBoxAreaDensity(Feature3DOrientedMinimumBoundingBox):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - oriented minimum bounding box area density"
        self.abbr_name = "morph_area_dens_ombb"
        self.ibsi_id = "IQYR"
        self.ibsi_compliant = False

    @staticmethod
    def _compute(data: Data3DOrientedMinimumBoundingBox, **kwargs) -> float:
        return data.volume / data.bounding_box_volume


class Feature3DMorphMajorAxisLength(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - major axis length"
        self.abbr_name = "morph_pca_maj_axis"
        self.ibsi_id = "TDIC"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        return data.semi_axes[2] * 2.0


class Feature3DMorphMinorAxisLength(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - minor axis length"
        self.abbr_name = "morph_pca_min_axis"
        self.ibsi_id = "P9VJ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        return data.semi_axes[1] * 2.0


class Feature3DMorphShortestAxisLength(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - shortest axis length"
        self.abbr_name = "morph_pca_least_axis"
        self.ibsi_id = "7J51"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        return data.semi_axes[0] * 2.0


class Feature3DMorphElongation(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - elongation"
        self.abbr_name = "morph_pca_elongation"
        self.ibsi_id = "Q3CK"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        return data.semi_axes[1] / data.semi_axes[2]


class Feature3DMorphFlatness(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - flatness"
        self.abbr_name = "morph_pca_flatness"
        self.ibsi_id = "N17B"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        return data.semi_axes[0] / data.semi_axes[2]


class Feature3DMorphApproximateEnclosingEllipsoidVolumeDensity(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - approximate enclosing ellipsoid volume density"
        self.abbr_name = "morph_vol_dens_aee"
        self.ibsi_id = "6BDE"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        if np.any(np.equal(data.semi_axes, 0.0)):
            return np.nan
        return 3.0 * data.volume / (4.0 * np.pi * np.prod(data.semi_axes))


class Feature3DMorphApproximateEnclosingEllipsoidAreaDensity(Feature3DPCA):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - approximate enclosing ellipsoid area density"
        self.abbr_name = "morph_area_dens_aee"
        self.ibsi_id = "RDD2"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DPrincipleComponents, **kwargs) -> float:
        if np.any(np.equal(data.semi_axes, 0.0)):
            return np.nan
        return data.area / data.get_ellipsoid_surface_area(n_degree=20)


class Feature3DMorphMinimumEnvelopingEllipsoidVolumeDensity(Feature3DMinimumEnvelopingEllipsoid):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - minimum volume enclosing ellipsoid volume density"
        self.abbr_name = "morph_vol_dens_mvee"
        self.ibsi_id = "SWZ1"
        self.ibsi_compliant = False

    @staticmethod
    def _compute(data: Data3DMinimumEnvelopingEllipsoid, **kwargs) -> float:
        if np.any(np.equal(data.semi_axes, 0.0)):
            return np.nan
        return 3.0 * data.volume / (4.0 * np.pi * np.prod(data.semi_axes))


class Feature3DMorphMinimumEnvelopingEllipsoidAreaDensity(Feature3DMinimumEnvelopingEllipsoid):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - minimum volume enclosing ellipsoid area density"
        self.abbr_name = "morph_area_dens_mvee"
        self.ibsi_id = "BRI8"
        self.ibsi_compliant = False

    @staticmethod
    def _compute(data: Data3DMinimumEnvelopingEllipsoid, **kwargs) -> float:
        if np.any(np.equal(data.semi_axes, 0.0)):
            return np.nan
        return data.area / data.get_ellipsoid_surface_area(n_degree=20)


class Feature3DMorphMoranIndex(Feature3DSpatial):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - Moran's I index"
        self.abbr_name = "morph_moran_i"
        self.ibsi_id = "N365"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DSpatial, **kwargs) -> float:
        return data.moran_i


class Feature3DMorphGearyMeasure(Feature3DSpatial):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Morphology (3D) - Geary's C measure"
        self.abbr_name = "morph_geary_c"
        self.ibsi_id = "NPT7"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: Data3DSpatial, **kwargs) -> float:
        return data.geary_c


def get_morphology_3d_class_dict() -> dict[str, Feature3DMorph]:
    class_dict = {
        "morph_volume": Feature3DMorphVolume,
        "morph_vol_approx": Feature3DMorphApproximateVolume,
        "morph_area_mesh": Feature3DMorphSurfaceArea,
        "morph_av": Feature3DMorphSurfaceVolumeRatio,
        "morph_comp_1": Feature3DMorphCompactness1,
        "morph_comp_2": Feature3DMorphCompactness2,
        "morph_sph_dispr": Feature3DMorphSphericalDisproportion,
        "morph_sphericity": Feature3DMorphSphericity,
        "morph_asphericity": Feature3DMorphAsphericity,
        "morph_com": Feature3DMorphCentreOfMassShift,
        "morph_integ_int": Feature3DMorphIntegratedIntensity,
        "morph_diam": Feature3DMorphMaximum3DDiameter,
        "morph_vol_dens_conv_hull": Feature3DMorphConvexHullVolumeDensity,
        "morph_area_dens_conv_hull": Feature3DMorphConvexHullAreaDensity,
        "morph_vol_dens_aabb": Feature3DMorphAxisAlignedBoundingBoxVolumeDensity,
        "morph_area_dens_aabb": Feature3DMorphAxisAlignedBoundingBoxAreaDensity,
        "morph_vol_dens_ombb": Feature3DMorphOrientedMinimumBoundingBoxVolumeDensity,
        "morph_area_dens_ombb": Feature3DMorphOrientedMinimumBoundingBoxAreaDensity,
        "morph_vol_dens_mvee": Feature3DMorphMinimumEnvelopingEllipsoidVolumeDensity,
        "morph_area_dens_mvee": Feature3DMorphMinimumEnvelopingEllipsoidAreaDensity,
        "morph_pca_maj_axis": Feature3DMorphMajorAxisLength,
        "morph_pca_min_axis": Feature3DMorphMinorAxisLength,
        "morph_pca_least_axis": Feature3DMorphShortestAxisLength,
        "morph_pca_elongation": Feature3DMorphElongation,
        "morph_pca_flatness": Feature3DMorphFlatness,
        "morph_vol_dens_aee": Feature3DMorphApproximateEnclosingEllipsoidVolumeDensity,
        "morph_area_dens_aee": Feature3DMorphApproximateEnclosingEllipsoidAreaDensity,
        "morph_moran_i": Feature3DMorphMoranIndex,
        "morph_geary_c": Feature3DMorphGearyMeasure
    }

    return class_dict


def generate_morph_3d_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[Feature3DMorph, None, None]:
    class_dict = get_morphology_3d_class_dict()
    morph_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_morphology_family():
        features = morph_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only morphological features, and return if none are present.
    features = [feature for feature in features if feature in morph_features]
    if len(features) == 0:
        return

    for feature in features:
        yield class_dict[feature](allow_approximation = not settings.no_approximation)



import copy
import warnings

import numpy as np
import pydicom

from mirp._data_import.mask_contour import ContourClass
from mirp.data_import.import_image import ImageFile
from mirp._data_import.dicom_file import MaskDicomFile
from mirp._masks.base_mask import BaseMask
from mirp._data_import.utilities import get_pydicom_meta_tag, has_pydicom_meta_tag


class MaskDicomFileRTSTRUCT(MaskDicomFile):
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        # RTSTRUCT does not have its own image information.
        self.image_origin = None
        self.image_orientation = None
        self.image_spacing = None
        self.image_dimension = None

    def is_stackable(self, stack_images: str):
        return False

    def create(self):
        return self

    def _complete_image_origin(self, force=False, frame_id=None):
        return

    def _complete_image_orientation(self, force=False, frame_id=None):
        return

    def _complete_image_spacing(self, force=False, frame_id=None):
        return

    def _complete_image_dimensions(self, force=False):
        return

    def _complete_frame_of_reference_uid(self):
        if self.frame_of_reference_uid is None:
            self.load_metadata(limited=True)

            # Try to obtain a frame of reference UID
            if has_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0020, 0x0052)):
                if get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0020, 0x0052), tag_type="str") is not None:
                    self.frame_of_reference_uid = get_pydicom_meta_tag(
                        dcm_seq=self.image_metadata,
                        tag=(0x0020, 0x0052),
                        tag_type="str"
                    )

            # For RT structure sets, the FOR UID may be tucked away in the Structure Set ROI Sequence
            if self.frame_of_reference_uid is None and \
                    has_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x3006, 0x0020)):
                structure_set_roi_sequence = self.image_metadata[(0x3006, 0x0020)]

                for structure_set_roi_element in structure_set_roi_sequence:
                    if has_pydicom_meta_tag(dcm_seq=structure_set_roi_element, tag=(0x3006, 0x0024)):
                        self.frame_of_reference_uid = get_pydicom_meta_tag(
                            dcm_seq=structure_set_roi_element,
                            tag=(0x3006, 0x0024),
                            tag_type="str"
                        )
                        break

    def check_mask(self, raise_error=True):
        if self.image_metadata is None:
            raise TypeError("DEV: the image_meta_data attribute has not been set.")

        if not get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x3006, 0x0020), test_tag=True):
            if raise_error:
                warnings.warn(
                    f"The current RT-structure set did not contain any ROI sequences. [{self.describe_self()}]",
                )
            return False

        roi_name_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_structure_set_roi_sequence, tag=(0x3006, 0x0026), tag_type="str", default=None)
            for current_structure_set_roi_sequence in self.image_metadata[(0x3006, 0x0020)]
        ]

        if len(roi_name_present) == 0 and raise_error:
            warnings.warn(
                f"The current RT-structure set does not contain any ROI contours. [{self.describe_self()}]"
            )
            return False

        if any(x is None for x in roi_name_present) and raise_error:
            warnings.warn(
                f"The current RT-structure set lacks one or more ROI names. [{self.describe_self()}]"
            )
            return False

        return True

    def load_data(self, **kwargs):
        pass

    def to_object(
            self,
            image: None | ImageFile,
            **kwargs
    ) -> list[BaseMask] | None:

        if image is None:
            raise TypeError(
                f"Converting an RT-structure to a segmentation mask requires that the corresponding image is set. "
                f"No image was provided. [{self.describe_self()}]"
            )
        else:
            image.complete()

        self.load_metadata()
        self.set_object_metadata()
        if not self.check_mask():
            return None

        mask_list = []

        # Find which roi numbers (3006,0022) are associated with which roi names (3004,0024).
        roi_name_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_structure_set_roi_sequence, tag=(0x3006, 0x0026), tag_type="str", default=None)
            for current_structure_set_roi_sequence in self.image_metadata[(0x3006, 0x0020)]
        ]
        roi_number_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_structure_set_roi_sequence, tag=(0x3006, 0x0022), tag_type="int", default=None)
            for current_structure_set_roi_sequence in self.image_metadata[(0x3006, 0x0020)]
        ]

        # Identify user-provided roi names.
        provided_roi_names = None
        if isinstance(self.roi_name, str):
            provided_roi_names = [self.roi_name]
        elif isinstance(self.roi_name, list):
            provided_roi_names = self.roi_name
        elif isinstance(self.roi_name, dict):
            provided_roi_names = list(self.roi_name.keys())

        if provided_roi_names is not None:
            roi_number_present = [
                x for ii, x in enumerate(roi_number_present) if roi_name_present[ii] in provided_roi_names
            ]
            roi_name_present = [x for x in roi_name_present if x in provided_roi_names]

        if len(roi_number_present) == 0:
            return None

        # Obtain segmentation from ROI Contour Sequences (0x3006, 0x0039).
        for roi_contour_sequence in self.image_metadata[(0x3006, 0x0039)]:

            # Get the roi number of the current ROI Contour Sequence
            current_roi_number = get_pydicom_meta_tag(
                dcm_seq=roi_contour_sequence,
                tag=(0x3006, 0x0084),
                tag_type="int")

            if current_roi_number not in roi_number_present:
                continue

            # Determine strategy for converting contours to mask objects.
            use_reference_image, use_orientation, use_position = self._convert_contour_using_image(
                roi_contour_sequence=roi_contour_sequence,
                image=image
            )

            if not use_reference_image:
                # Set up image based on contours.
                image = self._create_image_from_contour(
                    roi_contour_sequence=roi_contour_sequence,
                    image=image,
                    use_orientation=use_orientation,
                    use_position=use_position
                )

                image_data = self.convert_contour_to_mask(
                    roi_contour_sequence=roi_contour_sequence,
                    image=image
                )
            else:
                # Use associated image directly.
                image_data = self.convert_contour_to_mask(
                    roi_contour_sequence=roi_contour_sequence,
                    image=image
                )

            if image_data is None:
                continue

            if not np.any(image_data):
                continue

            # Complete a copy of the current object.
            temp_mask_object = MaskDicomFile(
                image_data=image_data,
                image_origin=image.image_origin,
                image_spacing=image.image_spacing,
                image_dimensions=image.image_dimension,
                image_orientation=image.image_orientation
            )
            temp_mask_object.update_image_data()

            current_roi_name = [
                x for ii, x in enumerate(roi_name_present) if roi_number_present[ii] == current_roi_number
            ][0]

            if isinstance(self.roi_name, dict):
                current_roi_name = self.roi_name.get(current_roi_name)

            mask_list += [
                BaseMask(
                    roi_name=current_roi_name,
                    sample_name=self.sample_name,
                    image_modality=self.modality,
                    image_data=temp_mask_object.image_data,
                    image_spacing=temp_mask_object.image_spacing,
                    image_origin=temp_mask_object.image_origin,
                    image_orientation=temp_mask_object.image_orientation,
                    image_dimensions=temp_mask_object.image_dimension,
                    metadata=self.object_metadata
                )
            ]

        if len(mask_list) == 0:
            return None

        return mask_list

    def _convert_contour_using_image(
            self,
            roi_contour_sequence: pydicom.Dataset,
            image: None | ImageFile
    ) -> tuple[bool, bool, bool]:
        """
        Determine whether the image should be used as reference for creating the mask. In general, if the contours
        can be directly mapped to slices, the image can be used as reference.

        Parameters
        ----------
        roi_contour_sequence: pydicom.Dataset
            DICOM ROI Contour Sequence read using pydicom.
        image: ImageFile
            Reference image

        Returns
        -------
        bool
            Boolean value that determines whether the image can be used as reference for creating the segmentation
            mask (`True`) or an image should be generated based on contour data (`False`).
        """

        from mirp._data_import.dicom_file import ImageDicomFile
        from mirp._data_import.dicom_file_stack import ImageDicomFileStack

        # In absence of an image, attempt to generate a reference segmentation mask using the ROI contour data
        # directly.
        if image is None:
            return False, False, False

        # Get series instance UID
        reference_series_instance_uid = None
        if isinstance(image, ImageDicomFile) or isinstance(image, ImageDicomFileStack):
            reference_series_instance_uid = image.series_instance_uid

        series_uid_correct = False
        if reference_series_instance_uid is not None:
            series_uid_correct = self.series_instance_uid == reference_series_instance_uid

        # Get SOP instance UID.
        reference_sop_instance_uid = None
        if isinstance(image, ImageDicomFile) and image.sop_instance_uid is not None:
            reference_sop_instance_uid = [image.sop_instance_uid]
        elif isinstance(image, ImageDicomFileStack) and image.sop_instance_uid is not None:
            reference_sop_instance_uid = image.sop_instance_uid

        sop_uid_correct = False
        if reference_sop_instance_uid is not None:
            referenced_sop_uids = []

            # Extract the SOP instance UIDs from images that are referenced to from the contour sequence.
            for contour_sequence in roi_contour_sequence[(0x3006, 0x0040)]:
                if has_pydicom_meta_tag(dcm_seq=contour_sequence, tag=(0x3006, 0x0016)):
                    for contour_image_sequence in contour_sequence[(0x3006, 0x0016)]:
                        referenced_sop_uids += [
                            get_pydicom_meta_tag(
                                dcm_seq=contour_image_sequence,
                                tag=(0x0008, 0x1155),
                                tag_type="str"
                            )
                        ]

            # Extract the SOP instance UIDs from images that are referenced to from the referenced frame of reference
            # sequence.
            if has_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x3006, 0x0010)):
                if has_pydicom_meta_tag(dcm_seq=self.image_metadata[(0x3006, 0x0010)][0], tag=(0x3006, 0x0012)):
                    for rt_referenced_series_sequence in self.image_metadata[(0x3006, 0x0010)][0][(0x3006, 0x0012)][0][(0x3006, 0x0014)]:
                        for contour_image_sequence in rt_referenced_series_sequence[(0x3006, 0x0016)]:
                            referenced_sop_uids += [
                                get_pydicom_meta_tag(
                                    dcm_seq=contour_image_sequence,
                                    tag=(0x0008, 0x1155),
                                    tag_type="str"
                                )
                            ]

            referenced_sop_uids = [referenced_sop_uid for referenced_sop_uid in referenced_sop_uids if referenced_sop_uid is not None]
            if len(referenced_sop_uids) > 1:
                sop_uid_correct = set(reference_sop_instance_uid) >= set(referenced_sop_uids)

        # If the reference SOP UID matches those found in the RT structure set: use these.
        if sop_uid_correct:
            return True, True, True

        # From here on, the SOP reference UIDs do not match, and we need to figure out if the RT structures
        # still align. First, convert contours in the contour sequence to internal contour objects.
        contour_objects = self._collect_contours(roi_contour_sequence=roi_contour_sequence)
        if contour_objects is None:
            return True, True, True

        # Convert contour points (world space) to voxel space.
        contour_objects = [contour.to_voxel_coordinates(image=image) for contour in contour_objects]

        # Check that within each plane a) each z-coordinate is constant, and b) maps to an integer value (i.e. the
        # slice number in the reference image).
        orientation_correct = True
        position_correct = True
        for contour in contour_objects:
            for sub_contour in contour.contour:
                # Check position
                slice_position = sub_contour[0, 0]
                if not np.allclose(sub_contour[:, 0], slice_position):
                    orientation_correct = False
                if not np.isclose(np.round(slice_position), slice_position):
                    position_correct = False

                if not orientation_correct:
                    break

            if not orientation_correct:
                break

        # If the orientation is not correct, the position should be revised as well.
        if not orientation_correct:
            return False, False, False
        if not position_correct:
            return False, orientation_correct, position_correct

        # Check that, if a mask is spread across multiple slices, at least two slices are adjacent OR the contours
        # refer to the series instance UID or to the SOP instance UID.
        slice_position = [sub_contour[0, 0] for contour in contour_objects for sub_contour in contour.contour]
        slice_position = np.unique(slice_position)

        # In case only one slice is present, we should attempt to compute differences.
        if len(slice_position) == 1:
            return True, True, True

        # Check if all slices are non-adjacent.
        if not np.isclose(np.min(np.diff(slice_position)), 1.0) and not sop_uid_correct and not series_uid_correct:
            return False, True, False

        return True, True, True

    def _create_image_from_contour(
            self,
            roi_contour_sequence: pydicom.Dataset,
            image: ImageFile,
            use_orientation: bool,
            use_position: bool
    ) -> ImageFile:
        # Check if any image should be created.
        if use_orientation and use_position:
            return image

        # Local copy of image.
        image = image.copy()

        # Convert contours in the contour sequence to internal contour objects.
        contour_objects = self._collect_contours(roi_contour_sequence=roi_contour_sequence)

        if use_orientation:
            mask_orientation = copy.deepcopy(image.image_orientation)
        else:
            points = np.vstack([sub_contour for sub_contour in contour_objects[0].contour])
            svd = np.linalg.svd(points.T - np.mean(points.T, axis=1, keepdims=True))

            # Extract the left singular U vectors.
            mask_orientation = svd.U[:, (2, 1, 0)]

        # Check right-handedness of orientation. When the orientation matrix is right-handed
        # (i.e. conforms to a standard coordinate system), its determinant should be 1.0. If not, we need to flip
        # one component (preferably the z-component.)
        if not np.around(np.linalg.det(mask_orientation), 6) == 1.0:
            mask_orientation = np.matmul(
                mask_orientation, np.array([[-1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
            )

        # The orientation matrix is now right-handed. However, to correctly set the position of the origin, we need
        # to orient the orientation matrix so that a positive increase in voxel translates yields a positive
        # position. The orientation matrix can be oriented in 8 different ways, of which one yields the expected
        # result. That is, we need to find a 3 x 3 rotation matrix B so that AB (1, 1, 1)^T > (0, 0, 0)^T,
        # with A the current orientation matrix. We use the matrix product AB as the orientation matrix
        # in subsequent steps.

        # B takes one of eight forms, being the inverse of right-handed orientation matrix corresponding to each of
        # the 8 vertices of a regular cube.
        vertex_matrices = [
            # 1. (0, 0, 0)
            np.array([
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0]
            ]),
            # 2. (0, 0, 1)
            np.array([
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, -1.0, 0.0]
            ]),
            # 3. (0, 1, 0)
            np.array([
                [1.0, 0.0, 0.0],
                [0.0, 0.0, -1.0],
                [0.0, 1.0, 0.0]
            ]),
            # 4. (0, 1, 1)
            np.array([
                [1.0, 0.0, 0.0],
                [0.0, -1.0, 0.0],
                [0.0, 0.0, -1.0]
            ]),
            # 5. (1, 0, 0)
            np.array([
                [-1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0]
            ]),
            # 6. (1, 0, 1)
            np.array([
                [-1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, -1.0]
            ]),
            # 7. (1, 1, 0)
            np.array([
                [-1.0, 0.0, 0.0],
                [0.0, -1.0, 0.0],
                [0.0, 0.0, 1.0]
            ]),
            # 8. (1, 1, 0)
            np.array([
                [-1.0, 0.0, 0.0],
                [0.0, 0.0, -1.0],
                [0.0, -1.0, 0.0]
            ])
        ]
        for vertex_matrix in vertex_matrices:
            # Because the vertex matrices are orthonormal, the inverse of a vertex matrix V is its transpose, i.e.
            # V^-1 = V^T. Thus, B = V^T, and we test A V^T (1, 1, 1)^T > (0, 0, 0)^T
            position_vector = np.matmul(
                np.matmul(mask_orientation, np.transpose(vertex_matrix)),
                np.transpose(np.array([1.0, 1.0, 1.0]))
            )
            # If a positive increase in voxel units corresponds to a positive increase in pseudo-world space, we
            # have found the correct rotation for the orientation matrix.
            if np.all(position_vector > 0.0):
                mask_orientation = np.matmul(mask_orientation, np.transpose(vertex_matrix))
                break

        # Ensure that the orientation matrix is normalised, i.e. the L2 norm is 1.0.
        l2_norm = np.around(np.linalg.norm(mask_orientation, ord=2), decimals=6)
        if l2_norm != 1.0:
            mask_orientation = mask_orientation / l2_norm

        if not use_position:
            # Spacing, origin and dimensions are actually easiest to define in voxel space, because that space is
            # orthonormal. If the mask space is positioned in an oblique manner, e.g. rotated in the x-y plane, the
            # world-space coordinates of contours do not directly translate into these characteristics. For conversion,
            # We use the orientation determined above, and a default origin of (0, 0, 0) and a unitary spacing, for
            # conversion to voxel coordinates.
            contours = [
                image.to_voxel_coordinates(
                    x=sub_contour.T,
                    origin=np.array([0.0, 0.0, 0.0]),
                    orientation=mask_orientation,
                    spacing=np.array([1.0, 1.0, 1.0])
                ).T
                for contour in contour_objects
                for sub_contour in contour.contour
            ]

            # Determine sample spacing. Note that because of unitary sample spacing during conversion from world space
            # to voxel space, the image spacing corresponds directly to world coordinates, e.g. a 1.0 voxel step in any
            # direction is a translation of 1.0 in world space in physical units.
            mask_z_spacing = np.min(np.diff(np.unique(np.vstack(contours)[:, 0])))
            mask_y_spacing = image.image_spacing[1]
            mask_x_spacing = image.image_spacing[2]
            mask_spacing = tuple([mask_z_spacing, mask_y_spacing, mask_x_spacing])

            # Determine origin. This is the translation with regard to the current origin in voxel space. This value is
            # then converted back to world space and used as the origin.
            mask_z_origin = np.min(np.unique(np.vstack(contours)[:, 0]))
            mask_y_origin = np.min(np.unique(np.vstack(contours)[:, 1])) - mask_y_spacing
            mask_x_origin = np.min(np.unique(np.vstack(contours)[:, 2])) - mask_x_spacing
            mask_origin = tuple(image.to_world_coordinates(
                x=np.array([mask_z_origin, mask_y_origin, mask_x_origin]),
                origin=np.array([0.0, 0.0, 0.0]),
                orientation=mask_orientation,
                spacing=np.array([1.0, 1.0, 1.0])
            ))

            # Determine dimensions.
            mask_z_dimension = int(np.ceil((np.max(np.unique(np.vstack(contours)[:, 0])) - mask_z_origin) / mask_z_spacing)) + 1
            mask_y_dimension = int(np.ceil((np.max(np.unique(np.vstack(contours)[:, 1])) - mask_y_origin) / mask_y_spacing)) + 1
            mask_x_dimension = int(np.ceil((np.max(np.unique(np.vstack(contours)[:, 2])) - mask_x_origin) / mask_x_spacing)) + 1
            mask_dimension = tuple([mask_z_dimension, mask_y_dimension, mask_x_dimension])
            
        else:
            mask_spacing = copy.deepcopy(image.image_spacing)
            mask_origin = copy.deepcopy(image.image_origin)
            mask_dimension = copy.deepcopy(image.image_dimension)

        # Update image
        image.image_spacing = mask_spacing
        image.image_origin = mask_origin
        image.image_dimension = mask_dimension
        image.image_orientation = mask_orientation

        return image

    def convert_contour_to_mask(
            self,
            roi_contour_sequence: pydicom.Dataset,
            image: ImageFile,
            draw_method: str = "ray_cast"
    ) -> None | np.ndarray:

        if image is None:
            return None

        # Initialise a list to contain contour data.
        contour_objects = self._collect_contours(roi_contour_sequence=roi_contour_sequence)
        if contour_objects is None:
            return None

        # Create an empty roi volume
        roi_mask = np.zeros(image.image_dimension, dtype=bool)

        # Create empty slice and mask lists.
        slice_list = []
        mask_list = []

        # Convert contour points (world space) to voxel space.
        contour_objects = [contour.to_voxel_coordinates(image=image) for contour in contour_objects]

        # Merge contours that belong to the same slice. Each contour should belong to a single slice, but multiple
        # contours of the same region of interest may be present in one slice. We therefore collect the contours for
        # each slice first.
        contour_objects = self._merge_contours(contour_objects)

        # Iterate over contours to fill out the mask
        for contour in contour_objects:

            # Ray casting method to draw segmentation map based on polygon contour - this is currently the only
            # implemented method.
            if draw_method == "ray_cast":
                contour_slice_list, contour_mask_list = contour.contour_to_grid_ray_cast(image=image)

                slice_list += contour_slice_list
                mask_list += contour_mask_list
            else:
                raise NotImplementedError(
                    f"The requested method ({draw_method}) for converting contours to a segmentation mask is not "
                    f"available. Currently available: ray_cast"
                )

        if len(slice_list) == 0:
            return None

        # Check for out-of-range slices.
        mask_list = [mask_list[ii] for ii, _ in enumerate(mask_list) if slice_list[ii] >= 0]
        slice_list = [slice_id for slice_id in slice_list if slice_id >= 0]

        if len(slice_list) == 0:
            return None

        # Identify any slices that lie outside the negative or positive z-range.
        mask_list = [mask_list[ii] for ii, _ in enumerate(mask_list) if abs(slice_list[ii]) < image.image_dimension[0]]
        slice_list = [slice_id for slice_id in slice_list if slice_id < image.image_dimension[0]]

        if len(slice_list) == 0:
            return None

        # Iterate over the elements in the slice list to set the mask.
        for ii in np.arange(len(slice_list)):
            slice_id = slice_list[ii]
            roi_mask[slice_id, :, :] = np.logical_or(roi_mask[slice_id, :, :], mask_list[ii])

        return roi_mask

    @staticmethod
    def _collect_contours(roi_contour_sequence: pydicom.Dataset) -> None | list[ContourClass]:
        contour_objects = []

        for contour_sequence in roi_contour_sequence[(0x3006, 0x0040)]:
            # Check if the geometric type exists (3006, 0042)
            if not get_pydicom_meta_tag(dcm_seq=contour_sequence, tag=(0x3006, 0x0042), test_tag=True):
                continue

            # Check if the geometric type equals "CLOSED_PLANAR"
            if get_pydicom_meta_tag(
                    dcm_seq=contour_sequence,
                    tag=(0x3006, 0x0042),
                    tag_type="str"
            ) != "CLOSED_PLANAR":
                continue

            # Check if contour data exists (3006, 0050)
            if not get_pydicom_meta_tag(dcm_seq=contour_sequence, tag=(0x3006, 0x0050), test_tag=True):
                continue

            # Read contour data.
            contour_data = np.array(
                get_pydicom_meta_tag(dcm_seq=contour_sequence, tag=(0x3006, 0x0050), tag_type="mult_float"),
                dtype=np.float64)
            contour_data = contour_data.reshape((-1, 3))

            # Determine if there is an offset (3006, 0045)
            contour_offset = np.array(
                get_pydicom_meta_tag(
                    dcm_seq=contour_sequence,
                    tag=(0x3006, 0x0045),
                    tag_type="mult_float",
                    default=[0.0, 0.0, 0.0]
                ),
                dtype=np.float64
            )

            # Remove the offset from the data
            contour_data -= contour_offset

            referenced_sop_uid = None
            if has_pydicom_meta_tag(dcm_seq=contour_sequence, tag=(0x3006, 0x0016)):
                for contour_image_sequence in contour_sequence[(0x3006, 0x0016)]:
                    referenced_sop_uid = get_pydicom_meta_tag(
                        dcm_seq=contour_image_sequence,
                        tag=(0x0008, 0x1155),
                        tag_type="str"
                    )

            # Store as contour.
            contour = ContourClass(
                contour=contour_data,
                reference_sop_uid=referenced_sop_uid
            )

            # Add contour data to the contour list
            contour_objects += [contour]

        if len(contour_objects) == 0:
            return None

        return contour_objects

    @staticmethod
    def _merge_contours(contour_objects: list[ContourClass]):
        """This function collects contours from the same slice."""
        # Find slice ids for each contour object.
        slice_ids = [contour.which_slice() for contour in contour_objects]

        unique_slice_ids: list[int] = list(np.unique(np.concatenate(slice_ids)))

        merged_contour_list = []

        for slice_id in unique_slice_ids:
            parent_contour = None
            merging_contours = None
            for ii in range(len(contour_objects)):
                if slice_id not in slice_ids[ii]:
                    continue

                if parent_contour is None:
                    parent_contour = contour_objects[ii]
                elif merging_contours is None:
                    merging_contours = [contour_objects[ii]]
                else:
                    merging_contours += [contour_objects[ii]]

            merged_contour_list += [
                parent_contour.merge(other_contours=merging_contours, slice_id=slice_id)]

        return merged_contour_list

    # @staticmethod
    # def _match_slice_position(slice_position, known_position, image_spacing_z):
    #     # Match slice position of mask with any known slice position.
    #     img_slice_position = slice_position * image_spacing_z
    #     position_difference = np.around(np.abs(img_slice_position - known_position), 3)
    #
    #     # Check if there is any matching position.
    #     if np.any(position_difference == 0.0):
    #         int_slice_position = int(np.argwhere(position_difference == 0.0))
    #     else:
    #         int_slice_position = None
    #
    #     return int_slice_position

    def export_roi_labels(self):

        self.load_metadata()

        # Find which roi numbers (3006,0022) are associated with which roi names (3004,0024).
        labels = [
            get_pydicom_meta_tag(
                dcm_seq=current_structure_set_roi_sequence, tag=(0x3006, 0x0026), tag_type="str", default=None)
            for current_structure_set_roi_sequence in self.image_metadata[(0x3006, 0x0020)]
        ]

        n_labels = max([1, len(labels)])

        if len(labels) == 0:
            labels = [None]

        # Get general attributes.
        parent_attributes = self._get_export_attributes()

        # Add roi labels as attribute.
        attributes = [("roi_label", labels)]
        parent_attributes.update(dict(attributes))

        return parent_attributes



import copy
import numpy as np

from mirp._data_import.generic_file import ImageFile


class ContourClass:

    def __init__(self, contour: np.ndarray, reference_sop_uid: str | None):
        # Convert contour to internal vertices, i.e. from (x, y, z) to (z, y, x)
        vertices = np.zeros(contour.shape, dtype=np.float64)
        vertices[:, 0] = contour[:, 2]
        vertices[:, 1] = contour[:, 1]
        vertices[:, 2] = contour[:, 0]

        self.contour = [vertices]
        self.reference_sop_uid = reference_sop_uid
        self.representation = "world"

    def copy(self):
        return copy.deepcopy(self)

    def to_voxel_coordinates(self, image: ImageFile):

        if self.representation == "voxel":
            return self

        self.contour = [
            np.transpose(image.to_voxel_coordinates(np.transpose(contour)))
            for contour in self.contour
            if contour is not None
        ]

        self.representation = "voxel"

        return self

    def to_world_coordinates(self, image: ImageFile):

        if self.representation == "world":
            return self

        self.contour = [
            np.transpose(image.to_world_coordinates(np.transpose(contour)))
            for contour in self.contour
            if contour is not None
        ]

        self.representation = "world"

        return self

    def which_slice(self) -> list[int]:

        if not self.representation == "voxel":
            raise ValueError("Contours should be represented in voxel space, not world space. Contact the devs.")

        slice_id = list(np.unique(np.concatenate([
            np.rint(contour[:, 0]).astype(int)
            for contour in self.contour
            if contour is not None
        ])))

        return slice_id

    def merge(
            self,
            other_contours=None,
            slice_id: None | int = None):

        # This function helps flatten nested contours.
        def _contour_extractor(contour_object_list, slice_id: None | int):
            new_contour_list = []
            for contour_object in contour_object_list:
                for contour in contour_object.contour:
                    if contour is not None:
                        if slice_id is not None:
                            # In case a certain slice is selected, only keep contour data pertaining to that slice.
                            contour = contour[np.rint(contour[:, 0]) == float(slice_id), :]

                        new_contour_list += [contour]

            return new_contour_list

        new_contour = self.copy()

        if other_contours is None:
            new_contour.contour = _contour_extractor([new_contour], slice_id=slice_id)
        else:
            new_contour.contour = _contour_extractor([new_contour] + other_contours, slice_id=slice_id)

        return new_contour

    def contour_to_grid_ray_cast(self, image: ImageFile):

        if not self.representation == "voxel":
            raise ValueError("Contours should be represented in voxel space, not world space. Contact the devs.")

        if self.contour is None:
            return None, None

        contour_slice_id = self.which_slice()
        if len(contour_slice_id) == 0:
            return None, None

        # Initiate a slice list and a mask list
        slice_list = []
        mask_list = []

        for current_slice_id in contour_slice_id:
            vertices = []
            lines = []
            vertex_offset = 0

            # Extract vertices and lines across contours.
            for contour in self.contour:
                current_vertices = contour[np.rint(contour[:, 0]) == float(current_slice_id), :][:, (1, 2)]
                if len(current_vertices) == 0:
                    continue

                # Reduce numerical issues by rounding of vertex positions.
                vertices += [np.around(current_vertices, decimals=5)]

                current_lines = np.vstack((
                    [np.arange(0, current_vertices.shape[0])],
                    [np.arange(-1, current_vertices.shape[0] - 1)])
                ).transpose()

                # Ensure that the lines are self-referential within each contour, i.e. -1 should never point to the last
                # vertex, because that vertex may not belong to the same contour.
                current_lines[0, 1] = current_vertices.shape[0] - 1
                current_lines += vertex_offset

                lines += [current_lines]

                # Update vertex_offset to account for each set of contours.
                vertex_offset += current_vertices.shape[0]

            # Merge vertices and lines
            vertices = np.concatenate(vertices, axis=0)
            lines = np.concatenate(lines, axis=0)

            slice_list.append(current_slice_id)
            mask_list.append(poly2grid(
                verts=vertices,
                lines=lines,
                spacing=np.array([1.0, 1.0]),
                origin=np.array([0.0, 0.0]),
                shape=np.array([image.image_dimension[1], image.image_dimension[2]])
            ))

        return slice_list, mask_list


def mesh2grid(verts, faces, spacing, origin=None):  # pragma: no cover

    # Remove superfluous dimensions
    verts = np.squeeze(np.asarray(verts))

    # Determine triangle vertices
    vert_1 = verts[faces[:, 0], :]
    vert_2 = verts[faces[:, 1], :]
    vert_3 = verts[faces[:, 2], :]

    # Determine AABB
    min_pos = np.min(verts, axis=0)
    max_pos = np.max(verts, axis=0)

    # Set origin
    if origin is None:
        origin = min_pos - 0.25 * spacing

    # Determine voxel grid
    vox_grid_dims = np.ceil((max_pos - origin) / spacing).astype(int)
    vox_grid = np.zeros(vox_grid_dims, dtype=int)

    # Set ray direction
    ray_dir = np.array((1.0, 0.0, 0.0))
    vox_span = origin[0] + np.arange(0, vox_grid_dims[0]) * spacing[0]

    # Iterate over voxels in x and y plane
    for y_ind in np.arange(0, vox_grid_dims[1]):
        for x_ind in np.arange(0, vox_grid_dims[2]):

            # Set ray origin
            ray_origin = np.array((origin[0] - 1.0,
                                   origin[1] + y_ind * spacing[1],
                                   origin[2] + x_ind * spacing[2]))

            vox_col = np.zeros(np.shape(vox_span), dtype=int)

            # Set mask to test only likely mesh triangles (in-plane mesh coordinates contain in-plane ray coordinates)
            simplex_mask = np.logical_and(np.abs(np.sum(np.sign(np.vstack((vert_1[:, 1], vert_2[:, 1], vert_3[:, 1])) - ray_origin[1]), axis=0)) < 3,
                                          np.abs(np.sum(np.sign(np.vstack((vert_1[:, 2], vert_2[:, 2], vert_3[:, 2])) - ray_origin[2]), axis=0)) < 3)

            # Go to next iterator if mask is empty
            if np.sum(simplex_mask) == 0:
                continue

            # Find scalar parameter t
            t_scal = np.array(
                list(map(lambda i_sel: ray_triangle_intersect(ray_orig=ray_origin, ray_dir=ray_dir, vert_1=vert_1[i_sel, :],
                                                              vert_2=vert_2[i_sel, :], vert_3=vert_3[i_sel, :]),
                         np.squeeze(np.where(simplex_mask)))))

            # Remove invalid and redundant results
            t_scal = np.unique(t_scal[np.isfinite(t_scal)])
            if t_scal.size == 0:
                continue

            # Update vox_col based on t_scal
            for t_curr in t_scal:
                vox_col[vox_span > t_curr + ray_origin[0]] += 1

            # Voxels in the roi cross an uneven number of meshes from the origin
            vox_grid[:, y_ind, x_ind] += vox_col % 2

    return vox_grid.astype(dtype=bool)


def poly2grid(verts, lines, spacing, origin, shape):

    # Set up line vertices
    vert_1 = verts[lines[:, 0], :]
    vert_2 = verts[lines[:, 1], :]

    # Remove lines with length 0 and center on origin
    line_mask = np.sum(np.abs(vert_1 - vert_2), axis=1) > 0.0
    vert_1 = vert_1[line_mask] - origin
    vert_2 = vert_2[line_mask] - origin

    # Find extent of contours in x
    x_min_ind = int(np.max([np.floor(np.min(verts[:, 1]) / spacing[1]), 0.0]))
    x_max_ind = int(np.min([np.ceil(np.max(verts[:, 1]) / spacing[1]), shape[1] * 1.0]))

    # Set up voxel grid and y-span
    vox_grid = np.zeros(shape, dtype=int)
    vox_span = origin[0] + np.arange(0, shape[0]) * spacing[0]

    # Set ray origin and direction (starts at negative y, and travels towards positive y
    ray_origin = np.array([-1.0, 0.0])
    ray_dir = np.array([1.0, 0.0])

    for x_ind in np.arange(x_min_ind, x_max_ind):
        # Update ray origin
        ray_origin[1] = origin[1] + x_ind * spacing[1]

        # Scan both forward and backward to resolve points located on the polygon
        vox_col_frwd = np.zeros(np.shape(vox_span), dtype=int)
        vox_col_bkwd = np.zeros(np.shape(vox_span), dtype=int)

        # Find lines that are intersected by the ray
        ray_hit = np.sum(np.sign(np.vstack((vert_1[:, 1], vert_2[:, 1])) - ray_origin[1]), axis=0)

        # If the ray crosses a vertex, the sum of the sign is 0 when the ray does not hit a vertex point, and -1 or 1 when it does.
        # In the latter case, we only keep of the vertices for each hit.
        simplex_mask = np.logical_or(ray_hit == 0, ray_hit == 1)

        # Go to next iterator if mask is empty
        if np.sum(simplex_mask) == 0:
            continue

        # Determine the selected vertices
        selected_verts = np.squeeze(np.where(simplex_mask))

        # Find intercept
        t_scal = np.array([ray_line_intersect(ray_orig=ray_origin, ray_dir=ray_dir, vert_1=vert_1[ii, :], vert_2=vert_2[ii, :]) for ii in selected_verts])

        # Remove invalid results
        t_scal = t_scal[np.isfinite(t_scal)]
        if t_scal.size == 0:
            continue

        # Update vox_col based on t_scal
        for t_curr in t_scal:
            vox_col_frwd[vox_span > t_curr + ray_origin[0]] += 1
        for t_curr in t_scal:
            vox_col_bkwd[vox_span < t_curr + ray_origin[0]] += 1

        # Voxels in the roi cross an uneven number of meshes from the origin
        vox_grid[:, x_ind] += np.logical_and(vox_col_frwd % 2, vox_col_bkwd % 2)

    return vox_grid.astype(dtype=bool)


def ray_triangle_intersect(ray_orig, ray_dir, vert_1, vert_2, vert_3):  # pragma: no cover
    # Implementation of the Moeller-Trumbore intersection algorithm to determine intersection point
    # between ray and triangle
    # This point satisfies ray_orig + t * ray_dir = (1 - u -v) * vert_1 + u * vert_2 + v * vert_3
    # See DOI: 10.1145/1198555.1198746
    # Back facing triangles are allowed

    epsilon = 0.000001

    # Define triangle edges
    edge_1 = vert_2 - vert_1
    edge_2 = vert_3 - vert_1

    # Calculate determinant
    p_vec = np.cross(ray_dir, edge_2)
    det = np.dot(edge_1, p_vec)

    # If determinant is near zero, the ray lies in the plane of the triangle, or is parallel to the triangle
    if (det > -epsilon) and (det < epsilon):
        return np.nan

    # Calculate inverse of the determinant
    inv_det = 1.0 / det

    # Calculate displacement vector between ray_origin and vert_1
    t_vec = ray_orig - vert_1

    # Calculate scalar parameter u and test bound
    u_scal = np.dot(t_vec, p_vec) * inv_det
    if (u_scal < 0.0) or (u_scal > 1.0):
        return np.nan

    # Calculate scalar parameter v and test bound
    q_vec = np.cross(t_vec, edge_1)
    v_scal = np.dot(ray_dir, q_vec) * inv_det
    if (v_scal < 0.0) or (u_scal + v_scal > 1.0):
        return np.nan

    # Calculate scalar parameter t
    t_scal = np.dot(edge_2, q_vec) * inv_det
    if t_scal > epsilon:
        return t_scal
    else:
        return np.nan


def ray_line_intersect(
        ray_orig: np.ndarray,
        ray_dir: np.ndarray,
        vert_1: np.ndarray,
        vert_2: np.ndarray
):

    epsilon = 0.000001

    # Define edge
    edge_line = vert_1 - vert_2

    # Define ray vertices
    r_vert_1 = ray_orig
    r_vert_2 = ray_orig + ray_dir
    edge_ray = - ray_dir

    # Calculate determinant - if close to 0, lines are parallel and will not intersect
    det = edge_ray[0] * edge_line[1] - edge_ray[1] * edge_line[0]
    if (det > -epsilon) and (det < epsilon):
        return np.nan

    # Calculate inverse of the determinant
    inv_det = 1.0 / det

    # Calculate determinant
    a11 = r_vert_1[0] * r_vert_2[1] - r_vert_1[1] * r_vert_2[0]
    a21 = vert_1[0] * vert_2[1] - vert_1[1] * vert_2[0]

    # Solve for x
    a12 = edge_ray[0]
    a22 = edge_line[0]
    x = np.linalg.det(np.array([[a11, a12], [a21, a22]])) * inv_det

    # Solve for y
    b12 = edge_ray[1]
    b22 = edge_line[1]
    y = np.linalg.det(np.array([[a11, b12], [a21, b22]])) * inv_det

    t = np.array([x, y])

    # Check whether the solution falls within the line segment
    u1 = np.around(np.dot(edge_line, edge_line), 5)
    u2 = np.around(np.dot(edge_line, vert_1-t), 5)
    if (u2 / u1) < 0.0 or (u2 / u1) > 1.0:
        return np.nan

    # Return scalar length from ray origin
    t_scal = np.linalg.norm(ray_orig-t)
    return t_scal



from typing import Any

from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame
from mirp._data_import.dicom_file_mr import ImageDicomFileMR
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileMRDCE(ImageDicomFileMR):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class ImageDicomFileMRDCEMultiFrame(ImageDicomMultiFrame, ImageDicomFileMRDCE):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import LaplacianOfGaussianTransformedImage
from mirp._imagefilters.utilities import FilterSet2D, FilterSet3D
from mirp.settings.generic import SettingsClass
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import pool_voxel_grids


class LaplacianOfGaussianFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "L6PA"

        self.sigma: float | list[float] = settings.img_transform.log_sigma
        self.sigma_cutoff = settings.img_transform.log_sigma_truncate
        self.pooling_method = settings.img_transform.log_pooling_method
        self.mode = settings.img_transform.log_boundary_condition

        # Riesz transformation settings.
        self.riesz_order: None | list[int] | list[list[int]] = None
        self.riesz_steered: bool = False
        self.riesz_sigma: None | float | list[float] = None
        if settings.img_transform.has_riesz_filter(x=name):
            self.riesz_order = settings.img_transform.riesz_order

            if settings.img_transform.has_steered_riesz_filter(x=name):
                self.riesz_steered = True
                self.riesz_sigma = settings.img_transform.riesz_filter_tensor_sigma

            # Riesz transformed filters are not IBSI-compliant
            self.ibsi_compliant = False

    def generate_object(self, allow_pooling: bool = True):
        # Generator for transformation objects.
        sigma = copy.deepcopy(self.sigma)
        if not isinstance(sigma, list):
            sigma = [sigma]

        # Nest sigma.
        if not self.pooling_method == "none" and allow_pooling:
            sigma = [sigma]

        riesz_order = copy.deepcopy(self.riesz_order)
        if riesz_order is None:
            riesz_order = [None]
        elif not all(isinstance(riesz_order_set, list) for riesz_order_set in riesz_order):
            riesz_order = [riesz_order]

        riesz_sigma = copy.deepcopy(self.riesz_sigma)
        if not isinstance(riesz_sigma, list):
            riesz_sigma = [riesz_sigma]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_riesz_order in riesz_order:
            for current_riesz_sigma in riesz_sigma:
                for current_sigma in sigma:
                    filter_object = copy.deepcopy(self)
                    filter_object.sigma = current_sigma
                    filter_object.riesz_order = current_riesz_order
                    filter_object.riesz_sigma = current_riesz_sigma

                    yield filter_object

    def transform(self, image: GenericImage) -> LaplacianOfGaussianTransformedImage:
        # Create placeholder Laplacian-of-Gaussian response map.
        response_map = LaplacianOfGaussianTransformedImage(
            image_data=None,
            sigma_parameter=self.sigma,
            sigma_cutoff_parameter=self.sigma_cutoff,
            pooling_method=self.pooling_method,
            boundary_condition=self.mode,
            riesz_order=self.riesz_order,
            riesz_steering=self.riesz_steered,
            riesz_sigma_parameter=self.riesz_sigma,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Set response voxel grid.
        response_voxel_grid = None

        # Initialise iterator ii to avoid IDE warnings.
        ii = 0
        for ii, pooled_filter_object in enumerate(self.generate_object(allow_pooling=False)):
            # Generate transformed voxel grid.
            pooled_voxel_grid = pooled_filter_object.transform_grid(
                voxel_grid=image.get_voxel_grid(),
                spacing=np.array(image.image_spacing)
            )

            # Pool voxel grids.
            response_voxel_grid = pool_voxel_grids(
                x1=response_voxel_grid,
                x2=pooled_voxel_grid,
                pooling_method=self.pooling_method
            )

        if self.pooling_method == "mean":
            # Perform final pooling step for mean pooling.
            response_voxel_grid = np.divide(response_voxel_grid, ii + 1)

        # Set voxel grid.
        response_map.set_voxel_grid(voxel_grid=response_voxel_grid)

        return response_map

    def transform_grid(
            self,
            voxel_grid: np.ndarray,
            spacing: np.ndarray):

        # Update sigma to voxel units.
        sigma = np.divide(np.full(shape=3, fill_value=self.sigma), spacing)

        # Determine the size of the filter
        filter_size = 1 + 2 * np.floor(self.sigma_cutoff * sigma + 0.5)

        if self.separate_slices:
            # Set the number of dimensions.
            d = 2.0

            # Create the grid coordinates, with [0, 0, 0] in the center.
            y, x = np.mgrid[:filter_size[1], :filter_size[2]]
            y -= (filter_size[1] - 1.0) / 2.0
            x -= (filter_size[2] - 1.0) / 2.0

            # Compute the square of the norm.
            norm_2 = np.power(y, 2.0) + np.power(x, 2.0)

        else:
            # Set the number of dimensions.
            d = 3.0

            # Create the grid coordinates, with [0, 0, 0] in the center.
            z, y, x = np.mgrid[:filter_size[0], :filter_size[1], :filter_size[2]]
            z -= (filter_size[0] - 1.0) / 2.0
            y -= (filter_size[1] - 1.0) / 2.0
            x -= (filter_size[2] - 1.0) / 2.0

            # Compute the square of the norm.
            norm_2 = np.power(z, 2.0) + np.power(y, 2.0) + np.power(x, 2.0)

        # Set a single sigma value.
        sigma = np.max(sigma)

        # Compute the scale factor
        scale_factor = - 1.0 / sigma ** 2.0 * np.power(1.0 / np.sqrt(2.0 * np.pi * sigma ** 2), d) * (d - norm_2 /
                                                                                                      sigma ** 2.0)

        # Compute the exponent which determines filter width.
        width_factor = - norm_2 / (2.0 * sigma ** 2.0)

        # Compute the weights of the filter.
        filter_weights = np.multiply(scale_factor, np.exp(width_factor))

        if self.separate_slices:
            # Set filter weights and create a filter.
            log_filter = FilterSet2D(
                filter_weights,
                riesz_order=self.riesz_order,
                riesz_steered=self.riesz_steered,
                riesz_sigma=self.riesz_sigma)

            # Convolve laplacian of gaussian filter with the image.
            response_map = log_filter.convolve(
                voxel_grid=voxel_grid,
                mode=self.mode,
                response="real")

        else:
            # Set filter weights and create a filter.
            log_filter = FilterSet3D(
                filter_weights,
                riesz_order=self.riesz_order,
                riesz_steered=self.riesz_steered,
                riesz_sigma=self.riesz_sigma)

            # Convolve laplacian of gaussian filter with the image.
            response_map = log_filter.convolve(
                voxel_grid=voxel_grid,
                mode=self.mode,
                response="real")

        # Compute the convolution
        return response_map
